"""
Functions to select optimal stimulation conditions based on the estimated y (i.e., MEPs in MEP studies).
"""

import numpy as np


def get_fim_sample(fun, x, p):
    """
    Get Fisher Information Matrix of one single sample.

    Parameters
    ----------
    fun : callable
        Callable the fisher information matrix is calculated for. The sample is passed as the first argument.
    x : float
        Sample passed to function.
    p : dict
        Dictionary containing the parameter estimates. The keys are the parameter names of fun.

    Returns
    -------
    fim_matrix : np.ndarray of float
        (n_params, n_params) Fisher information matrix.
    """

    # read function arguments
    params = p.keys()

    # determine gradient of function w.r.t. parameters using forward approximation
    dfdp = np.zeros(len(params))

    for i, para in enumerate(params):
        # copy original params dict
        p_dp = dict()

        for pa in params:
            p_dp[pa] = p[pa]

        # perturb parameter
        dp = p[para] / 1000
        p_dp[para] = p[para] + dp

        # determine gradient with forward approximation
        dfdp[i] = (fun(x, **p_dp) - fun(x, **p)) / dp

    fim_matrix = np.outer(dfdp, dfdp)

    return fim_matrix


def get_det_fim(x, fun, p, fim_matrix):
    """
    Updates the Fisher Information Matrix and returns the negative determinant based on the sample x.
    It is a score how much information the additional sample yields.

    Parameters
    ----------
    fun : callable
        Callable defined in interval [0, 1].
    x : float
        Single sample location (interval [0, 1]).
    p : dict
        Dictionary containing the parameter estimates. The keys are the parameter names of fun.
    fim_matrix : np.ndarray of float
        (n_params, n_params) Fisher Information Matrix.

    Returns
    -------
    det : float
        Determinant of the Fisher Information Matrix after adding sample x.
    """
    fim_matrix_sample = fim_matrix + get_fim_sample(fun=fun, x=x, p=p)
    sign, logdet = np.linalg.slogdet(fim_matrix_sample)

    return -sign * np.exp(logdet)


def init_fim_matrix(fun, x, p):
    """
    Initializes the Fisher Information Matrix based on the samples given in x.

    Parameters
    ----------
    fun : callable
        Callable defined in interval [0, 1].
    x : np.ndarray of float
        Initial sample locations (interval [0, 1]).
    p : dict
        Dictionary containing the parameter estimates. The keys are the parameter names of fun.

    Returns
    -------
    fim_matrix : np.ndarray of float [n_params x n_params]
        Fisher Information Matrix.
    """
    fim_matrix = np.zeros((len(p), len(p)))

    for x_i in x:
        fim_matrix += get_fim_sample(fun=fun, x=x_i, p=p)

    return fim_matrix


def get_optimal_sample_fim(fun, p, x=None):
    """
    Determines optimal location of next sample by maximizing the determinant of the Fisher Information Matrix.

    Parameters
    ----------
    fun : callable
        Callable (interval [0, 1]).
    x : np.ndarray of float, optional
        Previous sample locations (interval [0, 1]).
    p : dict
        Dictionary containing the parameter estimates. The keys are the parameter names of fun.

    Returns
    -------
    x_opt : float
        Optimal location of next sample (interval [0, 1]).
    """
    # initialize fim matrix with initial samples
    if x is None:
        fim_matrix = np.zeros((len(p), len(p)))
    else:
        fim_matrix = init_fim_matrix(fun=fun, x=x, p=p)

    # run optimization
    # res = minimize(fun=get_det_fim,
    #                method="trust-constr",
    #                bounds=((0, 1),),
    #                x0=np.array([0.5]),
    #                args=(fun, p, fim_matrix),
    #                options={"finite_diff_rel_step": 0.05, "xtol": 0.01})

    # res = minimize(fun=get_det_fim,
    #                 method="SLSQP",
    #                 x0=np.array([0.5]),
    #                 bounds=((0, 1),),
    #                 args=(fun, p, fim_matrix),
    #                 options={'disp': None, "eps": 0.001, "ftol": 1e-12}) #,
    # return res.x[0]

    x_bf = np.linspace(0, 1, 200)
    det = np.zeros(len(x_bf))
    for i in range(len(x_bf)):
        det[i] = get_det_fim(x=x_bf[i], fun=fun, p=p, fim_matrix=fim_matrix)

    x_opt = x_bf[np.argmin(det)]

    return x_opt



"""
The hdf5_io subpackage provides utilities for reading and writing data in the HDF5 format,
as well as generating XDMF files for visualization of the data. It includes functions for
writing surface data, creating XDMF files for surfaces and fibers, overlaying data stored
in HDF5 files, and writing coordinates to an XDMF file for visualization. This subpackage
is primarily used for handling and visualizing data related to neuroimaging and brain
stimulation studies.
"""
from .hdf5_io import *
from .xdmf import *



"""
:py:class:`pynibs.congruence` holds the initial congruence factor implementation as described in Weise, Numssen, et al., 2020 [1]_.
This code is mainly stored here for reproducibility reasons. The current approach (Numssen et al., 2021; [2]_) uses
the :py:class:`pynibs.regression` methods.

References
----------
.. [1] Weise, K., Numssen, O., Thielscher, A., Hartwigsen, G., & Knösche, T. R. (2020).
   A novel approach to localize cortical TMS effects. Neuroimage, 209, 116486.
.. [2] Numssen, O., Zier, A. L., Thielscher, A., Hartwigsen, G., Knösche, T. R., & Weise, K. (2021).
   Efficient high-resolution TMS mapping of the human motor cortex by nonlinear regression. NeuroImage, 245, 118654.
"""
from .congruence import *
from .ext_metrics import *
from .stimulation_threshold import *



import numpy as np
from scipy import interpolate


def sfap_dip(z):
    dist = 1

    i = np.zeros(len(z))
    i[np.isclose(z, 0, atol=0.01)] = 1
    i[np.isclose(z, dist, atol=0.01)] = -1
    return i


def sfap(z, sigma_i=1.01, d=55*1e-6, alpha=0.5):
    """
    Single fibre propagating transmembrane current (second spatial derivative of transmembrane potential).

    S. D. Nandedkar and E. V. Stalberg,“Simulation of single musclefiber action potentials”
    Med. Biol. Eng. Comput., vol. 21, pp. 158–165, Mar.1983.

    J.  Duchene  and  J.-Y.  Hogrel,“A  model  of  EMG  generation,”
    IEEETrans. Biomed. Eng., vol. 47, no. 2, pp. 192–200, Feb. 2000

    Hamilton-Wright, A., & Stashuk, D. W. (2005).
    Physiologically based simulation of clinical EMG signals.
    IEEE Transactions on biomedical engineering, 52(2), 171-183.

    Parameters
    ----------
    t : ndarray of float [n_t]
        Time in (ms)
    sigma_i : float, optional, default: 1.01
        Intracellular conductivity in (S/m)
    d : float, optional, default: 55*1e-6
        Diameter of muscle fibre in (m)
    v : float, optional, default: 1
        Conduction velocity in (m/s)
    alpha : float, optional, default: 0.5
        Scaling factor to adjust length of AP

    Returns
    -------
    i : ndarray of float [n_t]
        Transmembrane current of muscle fibre
    """
    # z = v * t
    z[z<0] = 0
    i = (sigma_i * np.pi * d**2) * 96 * alpha**3 * z * np.exp(-alpha*z) * (alpha**2*z**2 - 6*alpha*z + 6)
    # i = (sigma_i * np.pi * d**2) * 96 * (z) * (6-6*(z) + (z)**2)*np.exp(-z)
    # i = (sigma_i * np.pi * d**2)/4 * 3072 * z * (z**2 - 3*z + 1.5) * np.exp(-2*z)

    return i


def dipole_potential(z, loc, response):
    """
    Returns dipole potential at given coordinates z (interpolates given dipole potential)
    """
    res = np.zeros(len(z))
    mask = np.logical_and(z > loc[0], z < loc[-1])
    f = interpolate.interp1d(loc, response)
    res[mask] = f(z[mask])

    return res


def create_electrode(l_x, l_z, n_x, n_z):
    """
    Creates electrode coordinates

    Parameters
    ----------
    l_x : float
        X-extension of electrode in mm
    l_z : float
        Z-extension of electrode in mm
    n_x : int
        Number of point electrode in x-direction
    n_z : int
        Number of point electrodes in z-direction

    Returns
    -------
    electrode_coords : ndarray of float [n_ele x 3]
        Coordinates of point electrodes (x, y, z)
    """

    electrode_coords = np.zeros((n_x*n_z, 3))

    i = 0
    dx = l_x/(n_x-1)
    dz = l_z/(n_z-1)

    for i_x in range(n_x):
        for i_z in range(n_z):
            electrode_coords[i, :] = np.array([-l_x/2 + i_x*dx, 0, -l_z/2 + i_z*dz])
            i += 1

    return electrode_coords


def create_muscle_coords(l_x, l_y, n_x, n_y, h):
    """
    Create x and y coordinates of muscle fibres in muscle

    Parameters
    ----------
    l_x : float
        X-extension of muscle in mm
    l_y : float
        Y-extension of muscle in mm
    n_x : int
        Number of muscle fibres in x-direction
    n_y : int
        Number of muscle fibres in y-direction
    h : float
        Offset of muscle from electrode plane in mm

    Returns
    -------
    muscle_coords : ndarray of float [n_muscle x 3]
        Coordinates of muscle fibres in x-y plane (x, y, z)
    """

    muscle_coords = np.zeros((n_x*n_y, 3))

    i = 0

    if n_x == 1:
        dx = 0
    else:
        dx = l_x/(n_x-1)

    if n_y == 1:
        dy = 0
    else:
        dy = l_y/(n_y-1)

    for i_x in range(n_x):
        for i_y in range(n_y):
            muscle_coords[i, :] = np.array([-l_x/2 + i_x*dx, h + i_y*dy, 0])
            i += 1

    return muscle_coords


def create_muscle_fibre(x0, y0, L, n_fibre):
    """
    Creates muscle fibre coordinates (in z-direction)

    Parameters
    ----------
    x0 : float
        X-location of muscle fibre
    y0 : float
        Y-location of muscle fibre
    L : float
        Length of muscle fibre
    n_fibre : float
        Number of discrete fibre elements

    Returns
    -------
    fibre_coords : ndarray of float [n_fibre x 3]
        Coordinates of muscle fibre in z-direction (x, y, z)
    """
    dz_fibre = L/(n_fibre-1)
    fibre_coords = np.hstack((np.ones((n_fibre, 1))*x0,
                              np.ones((n_fibre, 1))*y0,
                              (-L/2 + np.arange(n_fibre) * dz_fibre)[:, np.newaxis]))

    return fibre_coords


def create_signal_matrix(T, dt, fibre_coords, z_e, v):
    """
    Create signal matrix containing the travelling action potential on the fibre

    Parameters
    ----------
    T : float
        Total time
    dt : float
        Time step
    fibre_coords : ndarray of float [n_fibre x 3]
        Coordinates of muscle fibre in z-direction (x, y, z)
    z_e : float
        Location of action potential generation
    v : float
        Velocity of action potential

    Returns
    -------
    signal_matrix : ndarray of float [n_time x n_fibre]
        Signal matrix containing the action potential values for each time step in the rows
    """
    N_t = int(T/dt + 1)  # number of time-steps
    t = np.linspace(0, T, N_t)
    n_fibre = fibre_coords.shape[0]
    dz_fibre = np.abs(fibre_coords[0, 2] - fibre_coords[1, 2])
    signal_matrix = np.zeros((N_t, n_fibre))
    z_e_idx = np.argmin(np.abs(fibre_coords[:, 2] - z_e))
    z_r = fibre_coords[z_e_idx:, 2] - fibre_coords[-1, 2]
    z_l = fibre_coords[:z_e_idx, 2] - fibre_coords[z_e_idx, 2]  # np.arange(-(L/2 + z_e), 0, dz_fibre)

    for i in range(N_t):
        z_l += v*dt
        z_r += v*dt

#         ap_r = sfap_dip(z=z_r)
#         ap_l = sfap_dip(z=z_l)

        ap_r = sfap(t=z_r, v=1)
        ap_l = sfap(t=z_l, v=1)

        signal_matrix[i, :z_e_idx] = ap_l
        signal_matrix[i, z_e_idx:] = np.flip(ap_r)

    return signal_matrix, t, fibre_coords[:, 2]


def hermite_rodriguez_1st(t, tau0=0, tau=0, lam=0.002):
    """
    First order Hermite Rodriguez function to model surface MUAPs

    Parameters
    ----------
    t : ndarray of float [n_t]
        Time axis in s
    tau0 : float, optional, default: 0
        initial shift to ensure causality in s
    tau : float, optional, default: 0
        shift (firing time) in s
    lam : float, optional, default: 2
        Timescale in s

    Returns
    -------
    y : ndarray of float [n_t]
        Surface MUAP
    """

    return -(t-tau0-tau) * np.exp(-((t-tau0-tau)/lam)**2)


def weight_signal_matrix(signal_matrix, fn_imp, t, z):
    """
    Weight signal matrix with impulse response from single dipole at every location
    """
    imp = np.loadtxt(fn_imp)
    imp[:, 0:3] = imp[:, 0:3]*1000
    imp[:, 3] = imp[:, 3]/np.max(imp[:, 3])
    signal_matrix_weighted = np.zeros((signal_matrix.shape[0], signal_matrix.shape[1], signal_matrix.shape[1]))

    for i_t in range(signal_matrix.shape[0]):
        print(f"{i_t}/{signal_matrix.shape[0]}")
        for i_z, z_w in enumerate(z):
            ir_interp = dipole_potential(z=z-z_w, loc=imp[:, 2], response=imp[:, 3])
            signal_matrix_weighted[i_t, i_z, :] = signal_matrix[i_t, i_z] * ir_interp

    signal_matrix_weighted = np.sum(signal_matrix_weighted, axis=1)

    return signal_matrix_weighted


def create_sensor_matrix(electrode_coords, fibre_coords, sigma_r=1, sigma_z=1):
    """
    Create sensor matrix containing the inverse distances from the point electrodes to the fibre elements
    weighted by the anisotropy factor of the muscle tissue.

    Parameters
    ----------
    electrode_coords : ndarray of float [n_ele x 3]
        Coordinates of point electrodes (x, y, z)
    fibre_coords : ndarray of float [n_fibre x 3]
        Coordinates of muscle fibre in z-direction (x, y, z)
    sigma_r : float, optional, default: 1
        Radial conductivity of muscle
    sigma_z : float, optional, default: 1
        Axial conductivity of muscle along fibre

    Returns
    -------
    sensor_matrix : ndarray of float [n_fibre x n_ele]
        Sensor matrix containing the inverse distances weighted with the anisotropy of muscle tissue
    """
    sigma_factor = sigma_z/sigma_r

    sensor_matrix = np.zeros((fibre_coords.shape[0], electrode_coords.shape[0]))

    for i in range(electrode_coords.shape[0]):
        r_f = np.linalg.norm(electrode_coords[i, :2] - fibre_coords[:, :2], axis=1)

        # sensor_matrix[:, i] = 1/np.linalg.norm(fibre_coords-electrode_coords[i, :], axis=1)
        sensor_matrix[:, i] = 1/np.sqrt(sigma_factor*r_f**2 + (fibre_coords[:, 2]-electrode_coords[i, 2])**2)

    return sensor_matrix


def compute_signal(signal_matrix, sensor_matrix):
    """
    Determine average signal from one single muscle fibre on all point electrodes

    Parameters
    ----------
    signal_matrix : ndarray of float [n_time x n_fibre]
        Signal matrix containing the action potential values for each time step in the rows
    sensor_matrix : ndarray of float [n_fibre x n_ele]
        Sensor matrix containing the inverse distances weighted with the anisotropy of muscle tissue

    Returns
    -------
    signal : ndarray of float [n_time]
        Average signal detected all point electrodes
    """
    signal = np.mean(np.matmul(signal_matrix, sensor_matrix), axis=1)

    return signal


def calc_mep_wilson(firing_rate_in, t, Qvmax=900, Qmmax=300, q=8, Tmin=14, N=100, M0=42, lam=0.002, tau0=0.006):
    """
    Determine motor evoked potential from incoming firing rate

    Parameters
    ----------
    firing_rate_in : ndarray of float [n_t]
        Input firing rate from alpha motor neurons
    t : ndarray of float [n_t]
        Time axis in s
    Qvmax : float, optional, default: 900
        Max of incoming firing rate [1/s]
    Qmmax : float, optional, default: 300
        Max of MU firing rate [1/s]
    q : float, optional, default: 8
        Min firing rate of MU [1/s]
    Tmin : float, optional, default: 14
        Min MU threshold [1/s]
    N : float, optional, default: 100
        Number of MU
    M0 : float, optional, default: 42
        Scaling constant of MU amplitude [mV/s]
    lam : float, optional, default: 0.002
        MUAP timescale of first order Hermite Rodriguez function [s]
    tau0 : float, optional, default: 0.006
        Standard shift of MUAP to ensure causality [s]

    Returns
    -------
    mep : ndarray of float [n_t]
        Motor evoked potential at surface electrode
    """
    dt = t[1] - t[0]            # time step
    k = np.arange(N)+1          # MU index
    firing_rate_in_mat = np.repeat(firing_rate_in[np.newaxis, :], N, axis=0)

    # determine MU thresholds
    alpha = 1/N*np.log(Qvmax/Tmin)
    Tk = Tmin * np.exp(alpha*k)

    # determine MU amplitude
    Mk = M0 * np.exp(alpha*k)

    # determine spike rate of MUs from incoming spike rate
    kappak = (Qmmax - q) / (Qvmax - Tk)
    Qk = q + kappak[:, np.newaxis] * (firing_rate_in_mat - Tk[:, np.newaxis])
    Qk[firing_rate_in_mat < Tk[:, np.newaxis]] = 0

    # determine spike times of MUs by integrating the spike rates
    Qk_int = np.cumsum(Qk, axis=1) * dt

    spike_times = []
    for k in range(N):
        Qk_int_max = int(np.floor(Qk_int[k, -1]))

        if Qk_int_max > 0:
            spike_times.append(np.zeros(Qk_int_max))

            for j in range(Qk_int_max):
                spike_times[k][j] = t[np.where(Qk_int[k, :] > (j+1))[0][0]]
        else:
            spike_times.append([])

    # determine EMG signal by summing up all MUAPS at determined firing times
    mep = np.zeros(len(t))

    for k in range(N):
        if len(spike_times[k]) > 0:
            for tau in spike_times[k]:
                mep += Mk[k] * hermite_rodriguez_1st(t=t, tau0=tau0, tau=tau, lam=lam)

    return mep


import re
import json
import pylab
import warnings
import datetime
import numpy as np
import pandas as pd
from lmfit import Model
import scipy.io as spio
from scipy.signal import butter, lfilter
import matplotlib.pyplot as plt

import pynibs

from pynibs.expio.fit_funs import sigmoid, sigmoid4_log, linear, exp, exp0

try:
    import biosig
except ImportError:
    # print("WARNING: Please install biosig from pynibs/pkg/biosig folder!")
    pass
np.seterr(over="ignore")


def get_mep_virtual_subject_TVS(x, p1=-5.0818, p2=-2.4677, p3=3.6466, p4=0.42639, p5=1.6665,
                                mu_y_add=10 ** -5.0818, mu_y_mult=-0.9645334, mu_x_add=0.68827324,
                                sigma_y_add=1.4739 * 1e-6, k=0.39316, sigma2_y_mult=2.2759 * 1e-2,
                                sigma2_x_add=2.3671 * 1e-2,
                                subject_variability=False, trial_variability=True):
    print("get_mep_virtual_subject_TVS() got renamed to get_virtual_mep_tvs()")
    raise NotImplementedError


def get_virtual_mep_tvs(x, p1=-5.0818, p2=-2.4677, p3=3.6466, p4=0.42639, p5=1.6665,
                        mu_y_add=10 ** -5.0818, mu_y_mult=-0.9645334, mu_x_add=0.68827324,
                        sigma_y_add=1.4739 * 1e-6, k=0.39316, sigma2_y_mult=2.2759 * 1e-2,
                        sigma2_x_add=2.3671 * 1e-2,
                        subject_variability=False, trial_variability=True):
    """
    Creates random mep data using the 3 variability source model from [1]_.
    There are typos in the paper but the code seems to be correct.
    Originally from S. Goetz: https://github.com/sgoetzduke/Statistical-MEP-Model .
    Rewritten from MATLAB to Python by Konstantin Weise.

    Parameters
    ----------
    x : np.ndarray of float
        (n_x) Normalized stimulator intensities [0 ... 1].
    p1 : float, default: -5.0818
        First parameter of sigmoidal hilltype function.
    p2 : float, default: 4.5323
        Second parameter of sigmoidal hilltype function.
    p3 : float, default: 3.6466
        Third parameter of sigmoidal hilltype function.
    p4 : float, default: 0.42639
        Fourth parameter of sigmoidal hilltype function.
    p5 : float, default: 1.6665
        Fifth parameter of sigmoidal hilltype function.
    mu_y_add : float, default:  10**-5.0818
        Mean value of additive y variability source.
    mu_y_mult : float, default: -0.9645334
        Mean value of multiplicative y variability source.
    mu_x_add : float, default: -0.68827324
        Mean value of additive x variability source.
    sigma_y_add : float, default: 1.4739*1e-6
        Standard deviation of additive y variability source.
    k : float, default: 0.39316
        Shape parameter of generalized extreme value distribution.
    sigma2_y_mult : float, default: 2.2759*1e-2
        Standard deviation of multiplicative y variability source.
    sigma2_x_add : float, default: 2.3671*1e-2
        Standard deviation of additive x variability source.
    subject_variability : bool, default: False
        Choose if shape parameters from IO curve are sampled from random distributions to model subject variability.
        This does not influence the trial-to-trial variability.
    trial_variability : bool, default: True
        Enable or disable trial-to-trial variability. Disabling it will result in ideal IO curves w/o noise.

    Returns
    -------
    mep : np.ndarray of float
        (n_x) Motor evoked potential values in mV.

    References
    ----------
    .. [1] Goetz, S. M., Alavi, S. M., Deng, Z. D., & Peterchev, A. V. (2019).
       Statistical Model of Motor-Evoked Potentials.
       IEEE Transactions on Neural Systems and Rehabilitation Engineering, 27(8), 1539-1545.
    """
    x = x * 100
    n_x = len(x)
    mu_y_add = 10 ** p1

    # determine variabilities
    if trial_variability:
        e_y_mult = 10 ** np.random.normal(loc=mu_y_mult, scale=np.sqrt(sigma2_y_mult), size=n_x)
        e_x_add = 10 ** np.random.normal(loc=mu_x_add, scale=np.sqrt(sigma2_x_add), size=n_x)

        # determine generalized value distribution of additive y variability
        p_e_y_add = pynibs.generalized_extreme_value_distribution(x=np.linspace(5e-6, 1e-4, 100000),
                                                                  mu=mu_y_add, sigma=sigma_y_add, k=k)

        # sample from generalized value distribution to determine additive y variability
        e_y_add = np.random.choice(np.linspace(5e-6, 1e-4, 100000), p=p_e_y_add / np.sum(p_e_y_add), size=n_x)

    else:
        e_y_mult = np.ones(n_x) * 10 ** mu_y_mult
        e_x_add = np.ones(n_x) * 10 ** mu_x_add
        e_y_add = np.ones(n_x) * mu_y_add

    x_prime = 2.224 * 1e-16 + x - 10 ** p5 + e_x_add
    x_prime[x_prime < 0] = 2.224 * 1e-16

    mep = e_y_add + np.exp(np.log(10) * (e_y_mult - 7 + (p2 + 7) / (1 + (10 ** p3) / (x_prime ** (10 ** p4)))))
    mep = mep * 1000

    return mep


def get_mep_virtual_subject_DVS(x, x0=0.5, r=10, amp=1, sigma_x=0, sigma_y=0, y0=1e-2, seed=None, rng=None):
    """
    Creates random mep data using the 2 variability source model from Goetz et al. 2014 [1]_ together with a standard 3
    parametric sigmoid function.

    Parameters
    ----------
    x : np.ndarray of float
        (n_x) Normalized stimulator intensities [0 ... 1].
    x0 : float, default: 0.5
        Location of turning point sigmoidal function.
    r : float, default: 0.25
        Steepness of sigmoidal function.
    amp : float, default: 1.0
        Saturation amplitude of sigmoidal function.
    sigma_x : float, default: 0.1
        Standard deviation of additive x variability source.
    sigma_y : float, default: 0.1
        Standard deviation of additive y variability source.
    y0 : float, default: 1e-2
        y-offset.
    seed : int, optional
        Seed to use.
    rng : numpy._generator.Generator, optional
        An already initialized random number generator that will be used instead of intializing a new one with seed.

    Returns
    -------
    mep : np.ndarray of float
        (n_x) Motor evoked potential values

    References
    ----------
    .. [1] Goetz, S. M., Luber, B., Lisanby, S. H., & Peterchev, A. V. (2014).
       A novel model incorporating two variability sources for describing motor evoked potentials.
       Brain stimulation, 7(4), 541-552.
    """
    if not isinstance(x, np.ndarray):
        x = np.atleast_1d(x)
    if type(rng) is np.random._generator.Generator:
        default_rng = rng
    else:
        if type(rng) is not type(None):
            print("[get_mep_virtual_subject_DVS] Invalid argument: get_mep_virtual_subject_DVS. Using numpy default_rng"
                  f" with seed={seed}")

        default_rng = np.random.default_rng(seed=seed)

    if not isinstance(x, np.ndarray):
        x = np.array([x])

    n_x = len(x)

    if sigma_x > 0:
        e_x_add = default_rng.normal(loc=0, scale=sigma_x, size=n_x)
    else:
        e_x_add = np.zeros(n_x)

    if sigma_y > 0:
        e_y_add = default_rng.normal(loc=0, scale=sigma_y, size=n_x)
    else:
        e_y_add = np.zeros(n_x)

    x_pert = x + e_x_add

    mep = 10 ** (sigmoid4_log(x=x_pert, x0=x0, r=r, amp=amp, y0=y0) + e_y_add)

    return mep


class Mep:
    """
    Mep object.

    Attributes
    ----------
    fun : function
        Function type to fit data with (``pynibs.sigmoid`` / ``pynibs.exp`` / ``pynibs.linear``).
    fun_sig : function
        Best fitting equivalent sigmoidal function (added by ``self.add_sigmoidal_bestfit()``).
    popt : np.ndarray of float
        (N_para) Fitted optimal function parameters.
    popt_sig : np.ndarray of float
        (3) Best fitting parameters ``x0``, ``r``, and ``amp`` of equivalent sigmoidal function.
    copt :  np.ndarray of float
        (N_para, N_para) covariance matrix of fitted parameters.
    pstd : np.ndarray of float
        (N_para) Standard deviation of fitted parameters.
    fit : object instance
        Gmodel object instance of parameter fit.
    x_limits : np.ndarray of float
        (2) Minimal and maximal value of intensity data.
    y_limits : np.ndarray of float
        (2) Minimal and maximal value of mep data.
    mt : float
        Motor threshold (MEP > 50 uV), evaluated from fitted curve, added after fitting.
    """

    # TODO: implement list of already fitted function types and their indices in self.fit[] for plot function. \
    #       condition name as attribute
    def __init__(self, intensities, mep, intensity_min_threshold=None, mep_min_threshold=None):
        """
        Initialize ``Mep`` object.

        Parameters
        ----------
        intensities : np.ndarray of float
            (N_mep) Intensities of TMS stimulator corresponding to measured MEP amplitudes.
        mep : np.ndarray of float
            (N_mep) MEP amplitudes in [V] measured during TMS.
        intensity_min_threshold : float, optional
            Minimum user defined intensity (values below it will be filtered out).
        mep_min_threshold : float, optional
            Minimum user defined MEP amplitude (values below it will be filtered out).
        """

        self.intensities_orig = intensities
        self.mep_orig = mep
        self.cvar = []
        self.popt = []
        self.pstd = []
        self.fun = sigmoid
        self.fun_sig = sigmoid
        self.popt_sig = []
        self.fit = []
        self.best_fit_idx = None
        self.mt = None
        self.constraints = None

        if intensity_min_threshold is None:
            intensity_min_threshold = -np.inf

        if mep_min_threshold is None:
            mep_min_threshold = -np.inf

        # filter negative meps or too small intensities
        mmask = np.where(self.mep_orig >= mep_min_threshold)
        self.intensities = self.intensities_orig[mmask]  # + np.random.randn(intensities.size)*1e-6
        self.mep = self.mep_orig[mmask]

        imask = np.where(self.intensities >= intensity_min_threshold)
        self.intensities = self.intensities[imask]  # + np.random.randn(intensities.size)*1e-6
        self.mep = self.mep[imask]

        self.x_limits = np.array([np.min(self.intensities), np.max(self.intensities)])
        self.y_limits = np.array([np.min(self.mep), np.max(self.mep)])

        # sort data in ascending order
        idx_sort = np.argsort(self.intensities)
        self.intensities = self.intensities[idx_sort]
        self.mep = self.mep[idx_sort]

    def fit_mep_to_function(self, p0=None):
        """
        Fits MEP data to function. The algorithm tries to fit the function first to a sigmoid, then to an
        exponential and finally to a linear function.

        Parameters
        ----------
        p0: np.ndarray of float, optional
            Initial guess of parameter values.

        Notes
        -----
        Additional attributes:

        Mep.popt : np.ndarray of float
            (N_para) Fitted optimal function parameters.
        Mep.copt : np.ndarray of float
            (N_para, N_para) Covariance matrix of fitted parameters.
        Mep.pstd : np.ndarray of float
            (N_para) Standard deviation of fitted parameters.
        Mep.fun : function
            Function mep data was fitted with.
        Mep.fit : object instance
            Gmodel object instance of parameter fit.
        Mep.mt : float
            Motor threshold (MEP > 50 uV).
        """
        if p0 is None:
            p0 = []

        delta = 0.4

        try_to_fit = True
        i_try = 1
        while try_to_fit:
            # unsuccessful fit
            if i_try == 4:
                break

            if i_try == 1:
                self.fun = sigmoid
                gmodel = Model(self.fun)
                self.fit = gmodel.fit(self.mep, x=self.intensities, x0=p0[0], r=p0[1], amp=p0[2])

            if i_try == 2:
                self.fun = exp
                gmodel = Model(self.fun)
                self.fit = gmodel.fit(self.mep, x=self.intensities, x0=p0[0], r=p0[1], y0=p0[2])

            if i_try == 3:
                self.fun = linear
                gmodel = Model(self.fun)
                self.fit = gmodel.fit(self.mep,
                                      x=self.intensities,
                                      m=(np.max(self.mep) - np.min(self.mep)) /
                                        (np.max(self.intensities) - np.min(self.intensities)),
                                      n=0)
                i_try = i_try + 1

            print('Fitting data to {} function ...'.format(str(self.fun.__name__)))

            # filter out unseccussful fit
            if self.fit.covar is None:
                i_try = i_try + 1
                print('Unsuccessful fit ... trying next function!')
                continue

            # get names of arguments of function
            argnames = self.fun.__code__.co_varnames[1:self.fun.__code__.co_argcount]

            # read out optimal function parameters
            self.popt = []
            for i in range(len(argnames)):
                self.popt.append(self.fit.best_values[argnames[i]])

            self.popt = np.asarray(self.popt)
            self.cvar = self.fit.covar
            self.pstd = np.sqrt(np.diag(self.cvar))

            # re-fit if std of any parameter is > delta
            if ((self.pstd / self.popt) > delta).any() and i_try < 3:
                print(f'Unsuccessful fit ... relative STD of one parameter is > {delta * 100}% ... '
                      f'trying next function!')
                i_try = i_try + 1
                continue

            # successful fit
            if ((self.pstd / self.popt) <= delta).all():
                break

        # determine motor threshold considering a threshold of 50 uV
        self.calc_motor_threshold(threshold=0.05)

    def fit_mep_to_function_multistart(self, p0=None, constraints=None, fun=None):
        """
        Fits MEP data to function.

        Parameters
        ----------
        p0 :  np.ndarray of float, optional
            Initial guess of parameter values.
        constraints : dict, optional
            Dictionary with parameter names as keys and [min, max] values as constraints.
        fun : list of functions, optional
            Functions to incorporate in the multistart fit (e.g.
            [``pynibs.sigmoid``, ``pynibs.exp0``, ``pynibs.linear``]).

        Notes
        -----
        Add Attributes:

        Mep.popt : np.ndarray of float
            (N_para) Best fitted optimal function parameters.
        Mep.copt : np.ndarray of float
            (N_para, N_para) Covariance matrix of best fitted parameters.
        Mep.pstd : np.ndarray of float
            (N_para) Standard deviation of best fitted parameters.
        Mep.fun : function
            Function of best fit mep data was fitted with.
        Mep.fit : list of fit object instances
            Gmodel object instances of parameter fits.
        Mep.best_fit_idx : int
            Index of best function fit (``fit[best_fit_idx]``).
        Mep.constraints : dict
            Dictionary with parameter names as keys and [min, max] values as constraints.
        """
        if fun is None:
            fun = [sigmoid, exp0, linear]
        if p0 is None:
            p0 = []

        self.constraints = constraints

        self.fit = []
        p0_passed = []
        argnames = []

        # fun = [sigmoid, exp0, linear]
        # # fun = [sigmoid]
        pstd_mean = np.zeros(len(fun))
        aic = np.zeros(len(fun))

        for i in range(len(fun)):
            if fun[i] == sigmoid:
                p0_passed = p0
            elif fun[i] == exp:
                p0_passed = p0  # (p0[0], p0[1])
            elif fun[i] == exp0:
                p0_passed = (p0[0], p0[1])
            elif fun[i] == linear:
                p0_passed = ((np.max(self.mep) - np.min(self.mep)) /
                             (np.max(self.intensities) - np.min(self.intensities)),
                             1)

            print('Fitting data to {} function ...'.format(str(fun[i].__name__)))
            self.fit.append(self.run_fit_multistart(fun[i],
                                                    x=self.intensities,
                                                    y=self.mep,
                                                    p0=p0_passed,
                                                    constraints=constraints))

            if self.fit[i] is not None:
                if self.fit[i].covar is not None:
                    pstd_mean[i] = np.mean(np.sqrt(np.diag(self.fit[i].covar)))
                else:
                    pstd_mean[i] = None

                aic[i] = self.fit[i].aic
                print("\t> found optimal parameters")
            else:
                pstd_mean[i] = None
                aic[i] = np.inf
                print("\t> failed")

            # get names of arguments of function
            argnames.append(fun[i].__code__.co_varnames[1:fun[i].__code__.co_argcount])

        # determine best model by aic (Akaike information criterion)
        self.best_fit_idx = int(np.argmin(aic))

        print("====================================")
        print("> Best fit: {} function".format(str(fun[self.best_fit_idx].__name__)))
        print("\n")

        # read out optimal function parameters from best fit
        self.popt = []
        for i in range(len(argnames[self.best_fit_idx])):
            self.popt.append(self.fit[self.best_fit_idx].best_values[argnames[self.best_fit_idx][i]])

        self.popt = np.asarray(self.popt)
        self.cvar = np.asarray(self.fit[self.best_fit_idx].covar)
        self.pstd = np.sqrt(np.diag(self.cvar))
        self.fun = fun[self.best_fit_idx]

        # determine motor threshold considering a threshold of 50 uV
        self.calc_motor_threshold(threshold=0.05)

    def run_fit_multistart(self, fun, x, y, p0, constraints=None, verbose=False, n_multistart=100):
        """
        Run multistart approach to fit data to function. ``n_multistart`` optimizations are performed based on random
        variations of the initial guess parameters ``p0``. The fit with the lowest AIC (Akaike information criterion),
        i.e. best fit is returned as gmodel fit instance.

        Parameters
        ----------
        fun : function
            Function mep data has to be fitted with.
        x : np.ndarray of float
            (N_data) Independent variable the data is fitted on.
        y : np.ndarray of float
            (N_data) Dependent data the curve is fitted through.
        p0 : np.ndarray of float or list of float
            Initial guess of parameter values.
        constraints : dict, optional
            Dictionary with parameter names as keys and [min, max] values as constraints.
        verbose : bool, default: False
            Show output messages.
        n_multistart : int, default: 100
            Number of repeated optimizations with different starting points to perform.

        Returns
        -------
        fit : object instance
            Gmodel object instance of best parameter fit with lowest parameter variance.
        """

        argnames = fun.__code__.co_varnames[1:fun.__code__.co_argcount]
        gmodel = Model(fun)

        if constraints:
            for i in range(len(list(constraints.keys()))):
                gmodel.set_param_hint(list(constraints.keys())[i],
                                      value=(constraints[list(constraints.keys())[i]][0] +
                                             constraints[list(constraints.keys())[i]][1]) / 2,
                                      min=constraints[list(constraints.keys())[i]][0],
                                      max=constraints[list(constraints.keys())[i]][1])

        n_p = len(argnames)
        fit = []

        # run n_s curve fits with random starting points
        for i in range(n_multistart):

            params = dict()

            # generate random samples of between constraint min and max or start points between 0 ... 2*p0
            for j in range(n_p):
                if constraints and argnames[j] in list(constraints.keys()):
                    params[argnames[j]] = np.random.random_sample(1) \
                                          * (constraints[argnames[j]][1] - constraints[argnames[j]][0]) \
                                          + constraints[argnames[j]][0]
                else:
                    params[argnames[j]] = p0[j] * (1 + (2 * np.random.random_sample(1) - 1))

            # set underflow and overflow warnings to exceptions, to be able to catch them
            old_np_settings = np.seterr(under='raise', over='raise')
            try:
                if fun == sigmoid:
                    fit.append(gmodel.fit(y, x=x, x0=params['x0'], r=params['r'], amp=params['amp'], verbose=False))
                if fun == exp:
                    fit.append(gmodel.fit(y, x=x, x0=params['x0'], r=params['r'], y0=params['y0'], verbose=False))
                if fun == exp0:
                    fit.append(gmodel.fit(y, x=x, x0=params['x0'], r=params['r'], verbose=False))
                if fun == linear:
                    fit.append(gmodel.fit(y, x=x, m=params['m'], n=params['n'], verbose=False))

            except (ValueError, FloatingPointError):
                if verbose:
                    print('\t Single fit failed in multistart optimization ... skipping value!')
                fit.append(None)

            # reset warnings to previous state
            np.seterr(**old_np_settings)

        # select best fit with lowest aic (Akaike information criterion) and return it
        pstd = []
        aic = []
        for i in range(n_multistart):
            if fit[i] is not None:
                if fit[i].covar is None:
                    pstd.append(float('Inf'))
                    aic.append(float('Inf'))
                else:
                    pstd.append(np.mean(np.sqrt(np.diag(fit[i].covar))))
                    aic.append(fit[i].aic)
            else:
                pstd.append(float('Inf'))
                aic.append(float('Inf'))
        idx = int(np.argmin(np.asarray(aic)))

        return fit[idx]

    def calc_motor_threshold(self, threshold):
        """
        Determine motor threshold of stimulator depending on MEP threshold given in [mV].

        Parameters
        ----------
        threshold : float
            Threshold of MEP amplitude in [mV].

        Notes
        -----
        Add Attributes:

        Mep.mt : float
            Motor threshold for given MEP threshold.
        """

        self.mt = np.nan

        # sample MEP curve very fine in given range
        intensities = np.linspace(self.x_limits[0], self.x_limits[1], 1000)
        mep_curve = self.eval_opt(intensities)
        i_threshold_idx = np.where(mep_curve > threshold)[0]

        if i_threshold_idx.any():
            self.mt = intensities[i_threshold_idx[0]]

    def plot(self, label, sigma=3, plot_samples=True, show_plot=False, fname_plot='', ylim=None, ylabel=None,
             fontsize_axis=10, fontsize_legend=10, fontsize_label=10, fontsize_title=10, fun=None):
        """
        Plotting mep data and fitted curve together with uncertainties.
        If ``fun == None`, the optimal function is plotted.

        Parameters
        ----------
        label : str
            Plot title.
        sigma : float, default: 3
            Factor of standard deviations the uncertainty of the fit is plotted with.
        plot_samples : bool, default: True
            Plot sampling curves of the fit in the uncertainty interval.
        show_plot : bool, default: False
            Show or hide plot window (``TRUE`` / ``FALSE``).
        fname_plot : str, default: ''
            Filename of plot showing fitted data (with .png extension).
        ylim : list of float, optional
            (2) Min and max values of y-axis.
        fontsize_axis : int, default: 10
            Fontsize of axis numbers.
        fontsize_legend : int, default: 10
            Fontsize of Legend labels.
        fontsize_label : int, default: 10
            Fontsize of axis labels.
        fontsize_title : int, default: 10
            Fontsize of title.
        fun : str, default: None
            Which function to plot (None, ``sigmoid``, ``exp``, ``linear``).

        Returns
        -------
        <File> : .png file
            Plot of Mep data and fit (format: png).
        """
        p = []
        if show_plot or fname_plot:
            x_range = np.max(self.intensities) - np.min(self.intensities)
            x = np.linspace(np.min(self.intensities) - 0.0 * x_range, np.max(self.intensities) + 0.0 * x_range, 100)

            # plot random sampling curves
            if plot_samples:
                n_s = 100
                params = np.random.randn(n_s, self.popt.shape[0]) * self.pstd + self.popt
                for i in range(n_s):
                    pylab.plot(x, self.fun(x, *params[i, :]), 'k', linewidth=0.25)

            # plot raw data
            p.append(pylab.plot(self.intensities, self.mep, 'o', label='data', markersize=3))

            # plot fit if present
            if self.fit:
                if fun is not None:
                    assert fun in ['sigmoid', 'exp0', 'linear']
                    # assert order of self.fit is ['sigmoid', 'linear', 'exp0']
                    if fun == 'sigmoid':
                        argnames = sigmoid.__code__.co_varnames[1:sigmoid.__code__.co_argcount]
                        opt = []
                        for i in range(len(argnames)):
                            opt.append(self.fit[0].best_values[argnames[i]])

                        opt = np.asarray(opt)
                        y = sigmoid(x, *opt)
                    elif fun == 'exp0':
                        argnames = exp0.__code__.co_varnames[1:exp0.__code__.co_argcount]
                        opt = []
                        for i in range(len(argnames)):
                            opt.append(self.fit[1].best_values[argnames[i]])

                        opt = np.asarray(opt)
                        y = exp0(x, *opt)
                    elif fun == 'linear':
                        argnames = linear.__code__.co_varnames[1:linear.__code__.co_argcount]
                        opt = []
                        for i in range(len(argnames)):
                            opt.append(self.fit[2].best_values[argnames[i]])

                        opt = np.asarray(opt)
                        y = linear(x, *opt)
                    else:
                        raise ValueError("{} is unknown. One of [None, 'sigmoid', 'exp', 'linear'")
                else:
                    y = self.fun(x, *self.popt)
                p.append(pylab.plot(x, y, 'r', label='fit ({})'.format(self.fun.__name__)))
                if ylim is not None:
                    pylab.ylim(ylim[0], ylim[1])
                pylab.title(label, size=fontsize_title)

                if sigma != 0:
                    y_min, y_max = self.eval_uncertainties(x, sigma=sigma)
                    p.append(pylab.fill_between(x, y_max, y_min, facecolor=[0.8, 0.8, 0.8],
                                                interpolate=True, label='std'))

            pylab.grid(color=[0.6, 0.6, 0.6], linestyle='--', linewidth=0.25)
            pylab.legend(loc=2, fontsize=fontsize_legend)
            pylab.xticks(size=fontsize_axis)
            pylab.yticks(size=fontsize_axis)
            pylab.xlabel(r'Stimulator intensity [A/$\mu$s]', size=fontsize_label)

            if ylabel is not None:
                pylab.ylabel(ylabel, size=fontsize_label)

            if show_plot:
                pylab.show()

            if fname_plot:
                pylab.savefig(fname_plot, format='png', bbox_inches='tight', pad_inches=0.01 * 4, dpi=600)

            pylab.close()

    def eval_uncertainties(self, x, sigma=1):
        """
        Evaluating approximated uncertainty interval around fitted distribution.

        Parameters
        ----------
        x : np.ndarray of float
            (N_x) Function values where uncertainty is evaluated.
        sigma : float, default: 1
            Standard deviation of parameters taken into account when evaluating uncertainty interval.

        Returns
        -------
        y_min : np.ndarray of float
            (N_x) Lower bounds of y-values.
        y_max : np.ndarray of float
            (N_x) Upper bounds of y-values.
        """

        if not self.fit:
            raise Exception('Please fit function first before evaluating uncertainties!')

        p = [np.array([self.popt[i] - sigma * self.pstd[i], self.popt[i] + sigma * self.pstd[i]])
             for i in range(self.popt.shape[0])]

        para_combinations = pynibs.get_cartesian_product(p)

        y = np.zeros((x.shape[0], para_combinations.shape[0]))

        for i in range(para_combinations.shape[0]):
            y[:, i] = self.eval(x, para_combinations[i, :])

        y_min = np.min(y, axis=1)
        y_max = np.max(y, axis=1)

        return y_min, y_max

    def eval_opt(self, x):
        """
        Evaluating fitted function with optimal parameters in points x.

        Parameters
        ----------
        x : np.ndarray of float
            (N_x) Function arguments.

        Returns
        -------
        y : np.ndarray of float
            (N_x) Function values.
        """

        y = self.fun(x, *self.popt)
        return y

    def eval(self, x, p):
        """
        Evaluating fitted function with optimal parameters in points x.

        Parameters
        ----------
        x: np.ndarray of float
            (N_x) Function arguments.
        p: tuple of float
            Function parameters.

        Returns
        -------
        y: np.ndarray of float
            (N_x) Function values.
        """

        y = self.fun(x, *p)
        return y

    def eval_fun_sig(self, x, p):
        """
        Evaluating optimally fitted sigmoidal function with optimal parameters in points x.

        Parameters
        ----------
        x: np.ndarray of float
            (N_x) Function arguments.
        p: tuple of float
            Function parameters.

        Returns
        -------
        y: np.ndarray of float
            (N_x) Function values.
        """

        y = self.fun_sig(x, *p)
        return y


def read_biosig_emg_data(fn_data, include_first_trigger=False, type="cfs"):
    """
    Reads EMG data from a biosig file.

    Parameters
    ----------
    fn_data : str
        Path to the biosig file.
    include_first_trigger : bool, default: False
        Whether to include the first trigger event in the data (default: False).
    type : str, default: 'cfs'
        Type of the biosig file.

    Returns
    -------
    emg_data : np.ndarray
        (num_sweeps, num_channels, samples_per_sweep) EMG data with shape.
    time_diff_list : list
        Time differences between trigger events in seconds.
    num_sweeps : int
        Number of sweeps in the EMG data.
    num_channels : int
        Number of channels in the EMG data.
    samples_per_sweep : int
        Number of samples per sweep in the EMG data.
    sampling_rate : int
        Sampling rate of the EMG data.
    """
    try:
        import biosig
    except ImportError:
        ImportError("Please install biosig from pynibs/pkg/biosig folder!")
        return

    if type == "cfs":  # TODO: also move the TXT reader here?
        cfs_fn = fn_data
        cfs_header = json.loads(biosig.header(cfs_fn))
        cfs_emg = biosig.data(cfs_fn)

        num_sweeps = cfs_header["NumberOfSweeps"]
        num_channels = cfs_emg.shape[1]

        total_num_samples = cfs_header["NumberOfSamples"]
        samples_per_sweep = int(total_num_samples / num_sweeps)
        sampling_rate = int(cfs_header["Samplingrate"])

        # get timestamps
        tms_pulse_timedelta = datetime.timedelta()
        # get hour, minute and second
        time_mep_list = []
        time_diff_list = []
        trigger_event_idcs = []
        if include_first_trigger:
            trigger_event_idcs.append(0)
            time_diff_list.append(0)
            time_mep_list.append(
                datetime.datetime.strptime(
                    cfs_header["EVENT"][0]["TimeStamp"], '%Y-%b-%d %H:%M:%S'
                )
                -
                datetime.timedelta(
                    seconds=float(cfs_header["EVENT"][0]["POS"])
                )
            )

        # convert time string into integer
        for event in cfs_header["EVENT"]:
            date = datetime.datetime.strptime(event["TimeStamp"], '%Y-%b-%d %H:%M:%S')

            # we are interested in the tms pulse time, so add it to ts
            date += tms_pulse_timedelta
            time_mep_list.append(date)
            time_diff_list.append((date - time_mep_list[0]).total_seconds())

            # compute indices in data block corresponding to the events
            if event["TYP"] == "0x7ffe":
                trigger_event_idcs.append(
                    round(event["POS"] * cfs_header["Samplingrate"])
                )

        num_sweeps = min(num_sweeps, len(trigger_event_idcs))

        emg_data = np.zeros((num_sweeps, num_channels, samples_per_sweep), dtype=np.float32)

        for c_idx in range(num_channels):
            # Use emg data starting from the index of the first trigger event
            # assumptions:
            # - after an initial offset all emg data were captured consecutively
            # - the first emg data frame may be captured without an explicit TMS
            #   tigger (eg. by checking the "write to disk" option)
            # - if we had dropouts in between the emg data block (not just at the
            #   beginning) we would need to adhere to the entire trigger_event_indices
            #   list.
            emg_data[:, c_idx, :] = np.reshape(
                cfs_emg[trigger_event_idcs[0]:, c_idx],
                (num_sweeps, samples_per_sweep)
            )

        return emg_data, time_diff_list, num_sweeps, num_channels, samples_per_sweep, sampling_rate


def get_mep_elements(mep_fn, tms_pulse_time, drop_mep_idx=None, cfs_data_column=0, channels=None, time_format="delta",
                     plot=False, start_mep=18, end_mep=35):
    """
    Read EMG data from CED .cfs or .txt file and returns MEP amplitudes.

    Parameters
    ----------
    mep_fn : string
        path to .cfs-file or .txt file (Signal export).
    tms_pulse_time : float
        Time in [s] of TMS pulse as specified in signal.
    drop_mep_idx : List of int or None, optional
        Which MEPs to remove before matching.
    cfs_data_column : int or list of int, default: 0
        Column(s) of dataset in cfs file. +1 for .txt.
    channels : list of str, optional
        Channel names.
    time_format : str, default: "delta"
        Format of mep time stamps in time_mep_lst to return.

        * ``"delta"`` returns list of datetime.timedelta in seconds.
        * ``"hms"`` returns datetime.datetime(year, month, day, hour, minute, second, microsecond).

    plot : bool, default: False
        Plot MEPs.
    start_mep : float, default: 18
        Start of time frame after TMS pulse where p2p value is evaluated (in ms).
    end_mep : float, default: 35
        End of time frame after TMS pulse where p2p value is evaluated (in ms).

    Returns
    -------
    p2p_array : np.ndarray of float
        (N_stim) Peak to peak values of N sweeps.
    time_mep_lst : list of datetime.timedelta
        MEP-timestamps
    mep_raw_data : np.ndarray of float
        (N_channel, N_stim, N_samples) Raw (unfiltered) MEP data.
    mep_filt_data : np.ndarray of float
        (N_channel, N_stim, N_samples) Filtered MEP data (Butterworth lowpass filter).
    time : np.ndarray of float
        (N_samples) Time axis corresponding to MEP data.
    mep_onset_array : np.ndarray of float
        (S_samples) MEP onset after TMS pulse.
    """
    # convert pulse time to datetime object in case of "delta"
    if time_format == "delta":
        tms_pulse_timedelta = datetime.timedelta(milliseconds=tms_pulse_time * 1000)
    elif time_format == "hms":
        tms_pulse_timedelta = datetime.timedelta()
    else:
        raise NotImplementedError("Specified time_format not implemented yet...")

    if mep_fn.endswith('.cfs'):
        # get data from cfs file
        import biosig
        mep_raw_data_tmp = biosig.data(mep_fn)
        mep_raw_data_tmp = mep_raw_data_tmp[:, cfs_data_column]  # get first channel

        # get header from cfs file
        cfs_header = biosig.header(mep_fn)

        # get timestamps
        # get all indices of timestamps from cfs header
        ts_mep_lst = [timestamp.start() for timestamp in re.finditer('TimeStamp', cfs_header)]
        # get hour, minute and second
        time_mep_list = []
        # convert time string into integer
        for index in ts_mep_lst:
            hour = int(cfs_header[index + 26:index + 28])
            minute = int(cfs_header[index + 29:index + 31])
            second = int(cfs_header[index + 32:index + 34])
            # fix bug with second 60
            if second == 60:
                ts = datetime.datetime(1900, 1, 1, hour, minute, 59)
                ts += datetime.timedelta(seconds=1)
            else:
                ts = datetime.datetime(1900, 1, 1, hour, minute, second)

            # we are interested in the tms pulse time, so add it to ts
            ts += tms_pulse_timedelta
            time_mep_list.append(ts)

        if time_format == "delta":
            time_mep_list = [time_mep_list[i] - time_mep_list[0] for i in range(len(time_mep_list))]
        if time_format == "hms":
            pass

        # add first timestamp (not saved by signal) and shift other by isi
        time_mep_list = [datetime.timedelta(seconds=0)] + [t + time_mep_list[1] - time_mep_list[0] for t in
                                                           time_mep_list]

        # get peak-to-peak values
        # get the ratio of samples per sweep
        sweep_index = cfs_header.find('NumberOfSweeps')
        comma_index = cfs_header.find(',', sweep_index)
        n_sweeps = int(cfs_header[sweep_index + 18:comma_index])
        record_index = cfs_header.find('NumberOfRecords')
        comma_index = cfs_header.find(',', record_index)
        records = int(cfs_header[record_index + 19:comma_index])
        n_samples = int(records / n_sweeps)
        if not isinstance(n_samples, int):
            print('Warning: Number of samples is not an integer.')
            # TODO: Correct get_mep_elements() sample number check. This does not work as expected (from Ole)

        # reshape numpy array
        mep_raw_arr = np.zeros((len(cfs_data_column), n_sweeps, n_samples))

        for i in cfs_data_column:
            mep_raw_arr[i, :, :] = np.reshape(mep_raw_data_tmp[:, i], (n_sweeps, n_samples))

        sampling_rate = get_mep_sampling_rate(mep_fn)

    elif mep_fn.endswith('.mat'):
        mep_data = spio.loadmat(mep_fn, struct_as_record=False, squeeze_me=True)

        # find data
        for k in mep_data.keys():
            if isinstance(mep_data[k], spio.matlab.mio5_params.mat_struct):
                mep_data = mep_data[k].__dict__
                break

        n_samples = mep_data['points']
        mep_raw_arr = mep_data['values'].transpose(1, 2, 0)
        time_mep_list = [datetime.timedelta(seconds=f.__dict__['start']) for f in mep_data['frameinfo']]
        sampling_rate = get_mep_sampling_rate(mep_fn)

    elif mep_fn.endswith('.txt'):
        warnings.warn(".txt import is deprecated - use .mat or .cfs.", DeprecationWarning)
        print("Reading MEP from .txt file")
        # The Signal text export looks like this:
        #
        # "s"\t"ADC 0"\t"ADC 1"
        # 0.00000000\t-0.066681\t-0.047607
        # 0.00025000\t-0.066376\t-0.049286
        # 0.00050000\t-0.066528\t-0.056610
        #
        # "s"\t"ADC 0"\t"ADC 1"
        # 0.00000000\t-0.066681\t-0.047607
        # 0.00025000\t-0.066376\t-0.049286
        # 0.00050000\t-0.066528\t-0.056610
        #
        # With first column = time, second = 1st electrode, ...
        # This is an example of 2 sweeps, 3 samples each, sampling rate = 4000

        # Find number of samples per sweep
        pattern = '"s"'
        with open(mep_fn, 'r') as f:
            for line_nr, line in enumerate(f):
                print(f'{line_nr}: {line}')
                if pattern in line and line_nr > 0:
                    # find second occurance of "s" -> end of first sweep
                    n_samples = line_nr
                    print(f'{line_nr}: {line}')
                if line != '\n':
                    last_sample_time = line

        # extract time (first column) of last samples
        last_sample_time = float(last_sample_time[0:last_sample_time.find('\t')])

        # subtract 2 because first row is header ("s"\t"ADC 0"\t"ADC 1") and last row is blank
        n_samples = n_samples - 2

        df_mep = pd.read_csv(mep_fn,
                             delimiter="\t",
                             skip_blank_lines=True,
                             skiprows=lambda x: x % (n_samples + 2) == 0 and x > 0)

        n_sweeps = int(df_mep.shape[0] / n_samples)
        mep_raw_arr = np.zeros((len(cfs_data_column), n_sweeps, n_samples))

        for i in range(n_sweeps):
            mep_raw_arr[:, i, :] = df_mep.iloc[i * n_samples:(i + 1) * n_samples, 1:].transpose()

        # get sampling rate by dividing number of sweeps by timing
        sampling_rate = int(mep_raw_arr.shape[2] - 1) / last_sample_time

        # build time_mep_list
        # we only have information about the single mep timings, so let's assume signal sticks strictly to the protocol
        sample_len = last_sample_time + 1 / sampling_rate

        # TODO: The ISI is missing here, do we want to add it to the subject object?
        time_mep_list = [datetime.timedelta(seconds=i * sample_len) +
                         tms_pulse_timedelta for i in range(mep_raw_arr.shape[1])]

    else:
        raise ValueError("Unknown MEP file extension. Use .cfs or .txt.")

    # get peak to peak value of every sweep and plot results in mep/plots/channels
    if channels is None:
        channels = [str(i) for i in cfs_data_column]

    tmp = np.zeros((mep_raw_arr.shape[0], mep_raw_arr.shape[1], 3)).astype(object)
    for i_channel in range(mep_raw_arr.shape[0]):
        print(f"Calculating p2p values for channel: {channels[i_channel]}")

        for i_zap in range(mep_raw_arr.shape[1]):
            tmp[i_channel, i_zap, 0], \
                tmp[i_channel, i_zap, 1], \
                tmp[i_channel, i_zap, 2] = calc_p2p(sweep=mep_raw_arr[i_channel, i_zap, :],
                                                    tms_pulse_time=tms_pulse_time,
                                                    sampling_rate=sampling_rate,
                                                    fn_plot=None,
                                                    start_mep=start_mep,
                                                    end_mep=end_mep)

    p2p_arr = np.zeros((tmp.shape[0], tmp.shape[1]))
    mep_onset_arr = np.zeros((tmp.shape[0], tmp.shape[1]))
    mep_filt_arr = np.zeros(mep_raw_arr.shape)

    time = np.arange(mep_raw_arr.shape[2]) / sampling_rate

    for idx_channel in cfs_data_column:
        for i, t in enumerate(tmp[idx_channel, :, :]):
            p2p_arr[idx_channel, i] = tmp[idx_channel, i, 0]
            mep_onset_arr[idx_channel, i] = tmp[idx_channel, i, 2]
            mep_filt_arr[idx_channel, i, :] = tmp[idx_channel, i, 1]

    if time_format == "delta":
        time_mep_list = [time_mep_list[i] - time_mep_list[0] for i in range(len(time_mep_list))]
    elif time_format == "hms":
        pass

    # remove MEPs according to drop_mep_idx and reset time
    if drop_mep_idx is not None:
        p2p_arr = np.delete(p2p_arr, drop_mep_idx)
        mep_onset_arr = np.delete(mep_onset_arr, drop_mep_idx)
        time_mep_list = np.delete(time_mep_list, drop_mep_idx)

    keep_mep_idx = [i for i in range(mep_raw_arr.shape[1]) if i not in np.array(drop_mep_idx)]
    mep_raw_arr = mep_raw_arr[:, keep_mep_idx, :]
    mep_filt_arr = mep_filt_arr[:, keep_mep_idx, :]

    return p2p_arr, time_mep_list, mep_raw_arr, mep_filt_arr, time, mep_onset_arr


def calc_p2p_old_exp0(sweep, start_mep=None, end_mep=None, tms_pulse_time=None, sampling_rate=None):
    """
    Calc peak-to-peak values of an mep sweep.
    This version was probably used in the ancient times of experiment 0.

    Parameters
    ----------
    sweep :  np.ndarray of float
        (Nx1) Input curve.
    start_mep : None
        Not used.
    end_mep : None
        Not used.
    tms_pulse_time : None
        Not used.
    sampling_rate : None
        Not used.

    Returns
    -------
    p2p : float
        Peak-to-peak value of input curve.
    """
    warnings.warn(DeprecationWarning("Use calc_p2p(). calc_p2p_old_exp0 only used for data-reproducibility"))

    # Filter requirements.
    order = 6
    fs = 16000  # sample rate, Hz
    cutoff = 2000  # desired cutoff frequency of the filter, Hz

    # Get the filter coefficients so we can check its frequency response.
    # import matplotlib.pyplot as plt
    # b, a = butter_lowpass(cutoff, fs, order)
    #
    # # Plot the frequency response.
    # w, h = freqz(b, a, worN=8000)
    # plt.subplot(2, 1, 1)
    # plt.plot(0.5 * fs * w / np.pi, np.abs(h), 'b')
    # plt.plot(cutoff, 0.5 * np.sqrt(2), 'ko')
    # plt.axvline(cutoff, color='k')
    # plt.xlim(0, 0.5 * fs)
    # plt.title("Lowpass Filter Frequency Response")
    # plt.xlabel('Frequency [Hz]')
    # plt.grid()

    sweep_filt = butter_lowpass_filter(sweep, cutoff, fs, order)

    # get indices for max
    index_max_begin = np.argmin(sweep) + 40  # get TMS impulse # int(0.221 / 0.4 * sweep.size)
    index_max_begin = int(0.221 / 0.4 * sweep.size)

    index_max_end = sweep_filt.size  # int(0.234 / 0.4 * sweep.size) + 1
    if index_max_begin >= index_max_end:
        index_max_begin = index_max_end - 1
    # index_max_end = index_max_begin + end_mep

    # get maximum and max index
    sweep_max = np.amax(sweep_filt[index_max_begin:index_max_end])
    sweep_max_index = index_max_begin + np.argmax(sweep_filt[index_max_begin:index_max_end])

    # if list of indices then get last value
    if sweep_max_index.size > 1:
        sweep_max_index = sweep_max_index[0]

    # get minimum and mix index
    index_min_begin = sweep_max_index  # int(sweep_max_index + 0.002 / 0.4 * sweep_filt.size)
    index_min_end = sweep_max_index + 40  # int(sweep_max_index + 0.009 / 0.4 * sweep_filt.size) + 1

    # Using the same window as the max should make this more robust
    # index_min_begin = index_max_begi
    sweep_min = np.amin(sweep_filt[index_min_begin:index_min_end])

    return sweep_max - sweep_min


def calc_p2p_old_exp1(sweep, start_mep=18, end_mep=35, tms_pulse_time=None, sampling_rate=2000):
    """
    Calc peak-to-peak values of an mep sweep.
    This version was probably used for the first fits of experiment 1.

    Parameters
    ----------
    sweep :  np.ndarray of float
        (Nx1) Input curve.
    start_mep: float, default: 18
        Starttime in [ms] after TMS for MEP search window.
    end_mep: float, default: 35
        Endtime in [ms] after TMS for MEP search window.
    tms_pulse_time : None
        Not used.
    sampling_rate : int, default: 2000
        Sampling rate in Hz.

    Returns
    -------
    p2p : float
        Peak-to-peak value of input curve.
    """
    warnings.warn(DeprecationWarning("Use calc_p2p(). calc_p2p_old_exp1 only used for data-reproducibility"))

    # Compute start and stop idx according to sampling rate
    start_mep = int((start_mep / 1000.) * sampling_rate)
    end_mep = int((end_mep / 1000.) * sampling_rate)

    # Filter requirements.
    order = 6
    fs = 16000  # sample rate, Hz
    cutoff = 2000  # desired cutoff frequency of the filter, Hz

    sweep_filt = butter_lowpass_filter(sweep, cutoff, fs, order)

    # get indices for max
    index_max_begin = np.argmin(sweep) + start_mep  # get TMS impulse # int(0.221 / 0.4 * sweep.size)
    if index_max_begin >= sweep_filt.size:
        index_max_begin = sweep_filt.size - 1
    # index_max_end = sweep_filt.size  # int(0.234 / 0.4 * sweep.size) + 1
    index_max_end = index_max_begin + end_mep

    # get maximum and max index
    sweep_max = np.amax(sweep_filt[index_max_begin:index_max_end])

    # Using the same window as the max should make this more robust
    index_min_begin = index_max_begin
    index_min_end = index_max_end
    sweep_min = np.amin(sweep_filt[index_min_begin:index_min_end])

    return sweep_max - sweep_min


def calc_p2p(sweep, tms_pulse_time=.2, start_mep=20, end_mep=35, measurement_start_time=0, sampling_rate=4000,
             cutoff_freq=500, fn_plot=None):
    """
    Calc peak-to-peak values of and mep sweep.

    Parameters
    ----------
    sweep :  np.ndarray of float
        (Nx1) Input curve.
    tms_pulse_time : float, default: .2
        Onset time of TMS pulse trigger in [s].
    start_mep : int, default: 18
        Start of p2p search window after TMS pulse in [ms].
    end_mep : int, default: 35
        End of p2p search window after TMS pulse in [ms].
    measurement_start_time : float, default: 0
        Start time of the EMG measurement in [ms].
    sampling_rate : int, default: 2000
        Sampling rate in Hz.
    cutoff_freq : int, default: 500
        Desired cutoff frequency of the filter in Hz.
    fn_plot : str, optional
        Filename of sweep to plot (.png). If None, plot is omitted.

    Returns
    -------
    p2p : float
        Peak-to-peak value of input curve.
    sweep_filt :  np.ndarray of float
        Filtered input curve (Butter lowpass filter with specified cutoff_freq).
    onset : float
        MEP onset after tms_pulse_time.
    """

    def time_to_idx_conversion(t):
        return int((t - measurement_start_time) * sampling_rate / 1000)

    def idx_to_time_conversion(i):
        return i * 1000 / sampling_rate + measurement_start_time

    # Compute start and stop idx according to sampling rate
    if tms_pulse_time > 1:
        warnings.warn(f"Is tms_pulse_time={tms_pulse_time} really in seconds?")

    # Filter requirements.
    order = 6

    sweep_filt = butter_lowpass_filter(sweep, cutoff_freq, sampling_rate, order)

    # beginning of mep search window
    srch_win_start = time_to_idx_conversion(tms_pulse_time * 1000 + start_mep)  # get TMS impulse

    if srch_win_start >= sweep_filt.size:
        srch_win_start = sweep_filt.size - 1
    srch_win_end = time_to_idx_conversion(tms_pulse_time * 1000 + end_mep)

    # get maximum and max index
    sweep_max = np.amax(sweep_filt[srch_win_start:srch_win_end])

    # Using the same window as the max should make this more robust
    sweep_min = np.amin(sweep_filt[srch_win_start:srch_win_end])
    p2p = sweep_max - sweep_min

    # find onset in [s] of mep after tms pulse
    onset_max = np.argmax(sweep[srch_win_start:srch_win_end])
    onset_min = np.argmin(sweep[srch_win_start:srch_win_end])
    if onset_min < onset_max:
        onset_max = onset_min
    onset_max += srch_win_start
    onset_max /= sampling_rate
    onset_max -= tms_pulse_time

    if fn_plot is not None:
        timepoints = np.asarray([idx_to_time_conversion(i) for i in np.arange(len(sweep))], dtype=np.float32)
        sweep_min_idx = np.argmin(sweep_filt[srch_win_start:srch_win_end]) + srch_win_start
        sweep_max_idx = np.argmax(sweep_filt[srch_win_start:srch_win_end]) + srch_win_start

        timepoints /= 1000
        plt.figure(figsize=[4.07, 3.52])
        plt.plot(timepoints, sweep)
        plt.plot(timepoints, sweep_filt)
        plt.scatter(timepoints[sweep_min_idx], sweep_min, 15, color="r")
        plt.scatter(timepoints[sweep_max_idx], sweep_max, 15, color="r")
        plt.plot(timepoints, np.ones(len(timepoints)) * sweep_min, linestyle="--", color="r", linewidth=1)
        plt.plot(timepoints, np.ones(len(timepoints)) * sweep_max, linestyle="--", color="r", linewidth=1)
        plt.grid()
        plt.legend(["raw", "filtered", "p2p"], loc='upper right')

        plt.xlim([np.max((tms_pulse_time - 0.01, np.min(timepoints))),
                  np.min((timepoints[srch_win_end] + .1, np.max(timepoints)))])
        plt.ylim([-1.1 * np.abs(sweep_min), 1.1 * np.abs(sweep_max)])

        plt.xlabel("time in s", fontsize=11)
        plt.ylabel("MEP in mV", fontsize=11)
        plt.tight_layout()

        plt.savefig(fn_plot, dpi=300, transparent=True)
        plt.close()

    return p2p, sweep_filt, onset_max


def get_mep_sampling_rate(cfs_path):
    """
    Returns sampling rate [Hz] for CED Signal EMG data in .cfs, .mat or .txt file.

    The sampling rate is saved in the cfs header like this:

    .. code-block:: sh

       ``Samplingrate"\t: 3999.999810,\n``

    Parameters
    ----------
    cfs_path : str
        Path to .cfs file or .txt file.

    Returns
    -------
    float : sampling rate
    """
    if cfs_path.endswith('.cfs'):
        # get header from cfs file
        cfs_header = biosig.header(cfs_path)

        # get start and end idx of 'samplingrate'
        idx_a = re.search('Samplingrate', cfs_header)

        # get idx of first '\n' after 'samplingrate'
        idx_b = re.search(',\n', cfs_header[idx_a.end():])

        sr_start = idx_a.end() + 4  # 'Samplingrate"\t: '
        return float(cfs_header[sr_start:idx_b.start() + idx_a.end()])

    elif cfs_path.endswith('.mat'):
        mep_data = spio.loadmat(cfs_path, struct_as_record=False, squeeze_me=True)
        for k in mep_data.keys():
            if isinstance(mep_data[k], spio.matlab.mio5_params.mat_struct):
                mep_data = mep_data[k].__dict__
                break
        return 1 / mep_data['interval']

    elif cfs_path.endswith('.txt'):
        # search for end of first frame / sweep
        pattern = '"s"'
        with open(cfs_path, 'r') as f:
            for line_nr, line in enumerate(f):
                if pattern in line and line_nr > 0:
                    # find second occurance of "s" -> end of first sweep
                    n_samples = line_nr
                    break
                if line != '\n':
                    last_sample_time = line
        # extract time (first column) of last samples
        last_sample_time = float(last_sample_time[0:last_sample_time.find('\t')])

        # subtract 2 because first row is header ("s"\t"ADC 0"\t"ADC 1") and last row is blank
        n_samples -= 2

        # get sampling rate by dividing number of sweeps by timing
        return n_samples / last_sample_time
    else:
        raise ValueError(f"cfs_path={cfs_path} must be .csf or .txt filename.")


def butter_lowpass(cutoff, fs, order=5):
    """
    Setup Butter low-pass filter and return filter parameters.

    Parameters
    ----------
    cutoff : float
        Cutoff frequency in [Hz].
    fs : float
        Sampling frequency in [Hz].
    order : int, default: 5
        Filter order.

    Returns
    -------
    b, a : np.ndarray, np.ndarray
        Numerator (b) and denominator (a) polynomials of the IIR filter.
    """

    nyq = 0.5 * fs
    normal_cutoff = cutoff / nyq
    b, a = butter(order, normal_cutoff, btype='low', analog=False)
    return b, a


def butter_lowpass_filter(data, cutoff, fs, order=5):
    """
    Applies Butterworth lowpass filter.

    Parameters
    ----------
    data : np.ndarray of float
        (N_samples) Input of the digital filter.
    cutoff : float
        Cutoff frequency in [Hz].
    fs : float
        Sampling frequency in [Hz].
    order : int
        Filter order.

    Returns
    -------
    y : np.ndarray
        (N_samples) Output of the digital filter.
    """

    b, a = butter_lowpass(cutoff, fs, order=order)
    y = lfilter(b, a, data)
    return y


def get_time_date(cfs_paths):
    """
    Get time and date of the start of the recording out of .cfs file.

    Parameters
    ----------
    cfs_paths : str
        Path to .cfs mep file.

    Returns
    -------
    time_date : str
        Date and time.
    """

    cfs_header = biosig.header(cfs_paths[0])
    index = cfs_header.find('StartOfRecording')
    time_date = cfs_header[index + 21:index + 40]
    return time_date


def scale_e_for_dvs(e, mt=60, upper_limit=100):
    """
    Scale the electric field for DVS and TVS to 0-1 range, with 0 being motor threshold.

    Parameters
    ----------
    e : np.array
        Electric field in V/m.
    mt : float, default: 60
        Desired motor threshold in V/m.
    upper_limit : float, default: 100
        Upper limit of the activation curve in V/m.

    Returns
    -------
    float : scaled electric field in V/m.
    """
    return (e - mt)/(upper_limit - mt)


"""
All functions to operate on TMS coils go here, for example to create ``.xdmf`` files to visualize coil positions.
"""
import os
import copy
import math
import h5py
import shutil
import random
import itertools
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial import Delaunay
from collections import OrderedDict

import pynibs


def get_coil_dipole_pos(coil_fn, matsimnibs):
    """
    Apply transformation to coil dipoles and return position.

    Parameters
    ----------
    coil_fn : str
        Filename of coil .ccd file.
    matsimnibs : np.ndarray of float
        Transformation matrix.

    Returns
    -------
    dipoles_pos : np.ndarray
        (N, 3) Cartesian coordinates (x, y, z) of coil magnetic dipoles.
    """
    if coil_fn[-3:] == "nii":
        coil_fn = coil_fn[:-3] + "ccd"
    coil_data = np.genfromtxt(coil_fn, delimiter=' ', skip_header=3)
    coil_dipoles = coil_data[:, 0:3]
    coil_dipoles *= 1e3
    coil_dipoles = np.hstack([coil_dipoles, np.ones((coil_dipoles.shape[0], 1))])

    return matsimnibs.dot(coil_dipoles.T).T[:, :3]


def check_coil_position(points, hull):
    """
    Check if magnetic dipoles are lying inside head region

    Parameters
    ----------
    points : np.ndarray of float
        (N_points, 3) Coordinates (x,y,z) of magnetic dipoles
    hull : Delaunay object or np.ndarray of float
        (N_surface_points, 3) Head surface data

    Returns
    -------
    valid : bool
        Validity of coil position:
        TRUE: valid
        FALSE: unvalid
    """
    # make Delaunay grid if not already passed
    if not isinstance(hull, Delaunay):
        hull = Delaunay(hull)

    # filter out points which are outside bounding box to save some time
    bounds = [np.min(hull.points), np.max(hull.points)]
    points_inside = np.logical_and((points > bounds[0]).all(axis=1),
                                   (points < bounds[1]).all(axis=1))

    # test if points are inside (True)
    inside = hull.find_simplex(points[points_inside]) >= 0
    valid = not (inside.any())

    return valid


def calc_coil_transformation_matrix(LOC_mean, ORI_mean, LOC_var, ORI_var, V):
    """
    Calculate the modified coil transformation matrix needed for SimNIBS based on location and orientation
    variations observed in the framework of uncertainty analysis

    Parameters
    ----------
    LOC_mean : np.ndarray of float
        (3), Mean location of TMS coil
    ORI_mean : np.ndarray of float
        (3 x 3) Mean orientations of TMS coil

        .. math::
            \\begin{bmatrix}
            | & | & | \\\\
            x & y & z \\\\
            | & | & | \\\\
            \\end{bmatrix}

    LOC_var : np.ndarray of float
        (3) Location variation in normalized space (dx', dy', dz'), i.e. zero mean and projected on principal axes
    ORI_var : np.ndarray of float
        (3) Orientation variation expressed in Euler angles [alpha, beta, gamma] in deg
    V : np.ndarray of float
        (3x3) V-matrix containing the eigenvectors from  _,_,V = numpy.linalg.svd

    Returns
    -------
    mat : np.ndarray of float
    (4, 4) Transformation matrix containing 3 axis and 1 location vector:

        .. math::
            \\begin{bmatrix}
            | & | & | &  |   \\\\
            x & y & z & pos  \\\\
            | & | & | &  |   \\\\
            0 & 0 & 0 &  1   \\\\
            \\end{bmatrix}
    """
    # calculate rotation matrix for angle variation (angle_var in deg)
    rotation_matrix = pynibs.euler_angles_to_rotation_matrix(ORI_var * np.pi / 180.)

    # determine new orientation
    ori = np.dot(ORI_mean, rotation_matrix)

    # determine new location by retransforming from normalized space to regular space
    locations = np.dot(LOC_var, V) + LOC_mean

    # concatenate results
    mat = np.hstack((ori, locations[:, np.newaxis]))
    mat = np.vstack((mat, [0, 0, 0, 1]))

    return mat


def calc_coil_position_pdf(fn_rescon=None, fn_simpos=None, fn_exp=None, orientation='quaternions',
                           folder_pdfplots=None):
    """
    Determines the probability density functions of the transformed coil position (x', y', z') and quaternions of
    the coil orientations (x'', y'', z'')

    Parameters
    ----------
    fn_rescon : str
        Filename of the results file from TMS experiments (results_conditions.csv)
    fn_simpos : str
        Filename of the positions and orientation from TMS experiments (simPos.csv)
    fn_exp : str
        Filename of experimental.csv file from experiments
    orientation: str
        Type of orientation estimation: 'quaternions' or 'euler'
    folder_pdfplots : str
        Folder, where the plots of the fitted pdfs are saved (omitted if not provided)

    Returns
    -------
    pdf_paras_location : list of list of np.ndarray
        [n_conditions] Pdf parameters (limits and shape) of the coil position for x', y', and z' for each:

            - beta_paras ... [p, q, a, b] (2 shape parameters and limits)
            - moments    ... [data_mean, data_std, beta_mean, beta_std]
            - p_value    ... p-value of the Kolmogorov Smirnov test
            - uni_paras  ... [a, b] (limits)
    pdf_paras_orientation_euler : list of np.ndarray
        [n_conditions] Pdf parameters (limits and shape) of the coil orientation Psi, Theta, and Phi for each:

            - beta_paras ... [p, q, a, b] (2 shape parameters and limits)
            - moments    ... [data_mean, data_std, beta_mean, beta_std]
            - p_value    ... p-value of the Kolmogorov Smirnov test
            - uni_paras  ... [a, b] (limits)
    OP_mean : List of [3 x 4] np.ndarray
        [n_conditions] List of mean coil position and orientation for different conditions (global coordinate system)

        .. math::
            \\begin{bmatrix}
            |  &   |   &   |   &  |   \\\\
            ori_x & ori_y & ori_z & pos  \\\\
            |  &   |   &   |   &  |   \\\\
            \\end{bmatrix}

    OP_zeromean : list of [3 x 4 x n_con_each] np.ndarray [n_conditions]
        List over conditions containing zero-mean coil orientations and positions
    V : list of [3 x 3] np.ndarrays [n_conditions]
        Transformation matrix of coil positions from global coordinate system to transformed coordinate system
    P_transform : list of np.ndarray [n_conditions]
        List over conditions containing transformed coil positions [x', y', z'] of all stimulations
        (zero-mean, rotated by SVD)
    quaternions : list of np.ndarray [n_conditions]
        List over conditions containing imaginary part of quaternions [x'', y'', z''] of all stimulations
    """
    import pygpc

    if folder_pdfplots is not None:
        make_pdf_plot = True
    else:
        make_pdf_plot = False
        folder_pdfplots = ''

    if fn_rescon and fn_simpos:
        positions_all, conditions, position_list, _, _ = pynibs.read_exp_stimulations(fn_rescon, fn_simpos)

        # sort POSITIONS according to CONDITIONS, idx_con is alphabetically sorted, first index of condition[*]
        # appeareance
        conditions_unique, idx_con, n_con_each = np.unique(conditions,
                                                           return_index=True,
                                                           return_inverse=False,
                                                           return_counts=True)

        idx_con_sort = np.argsort(idx_con)
        conditions_unique = conditions_unique[idx_con_sort]  # conditions_unique is ordered like experimental data
        n_con_each = n_con_each[idx_con_sort]
        n_condition = len(conditions_unique)
        n_zaps = len(positions_all)

        pos_idx = 0
        k = 0

        # Orientation and position tensors (list over conditions containing arrays of [3 x 4 x n_con_each])
        op = [np.zeros((3, 4, n_con_each[i])) for i in range(n_condition)]

        # read data from global position_list
        for i in range(n_condition):
            while position_list[pos_idx][len(position_list[0]) - 3] == conditions_unique[i]:
                op[i][:, :, k] = np.array(positions_all[pos_idx])[0:3, 0:4]
                pos_idx = pos_idx + 1
                k = k + 1
                if pos_idx == n_zaps:
                    break
            k = 0

        # determine matrix of mean location and orientation (list over conditions containing arrays of [3 x 4])
        op_mean = [np.mean(op[i], axis=2) for i in range(n_condition)]

    elif fn_exp:
        exp = pynibs.read_csv(fn_exp)
        exp_cond = pynibs.sort_by_condition(exp)

        conditions_unique = [exp_cond[i_cond]['condition'][0] for i_cond in range(len(exp_cond))]

        n_con_each = [len(e['condition']) for e in exp_cond]
        n_condition = len(exp_cond)

        # Orientation and position tensors (list over conditions containing arrays of [3 x 4 x n_con_each])
        op = [np.zeros((3, 4, n_con_each[i])) for i in range(n_condition)]

        for i in range(n_condition):
            for k in range(n_con_each[i]):
                op[i][:, :, k] = exp_cond[i]['coil_mean_matrix'][k][0:3, :]

        # determine matrix of mean location and orientation (list over conditions containing arrays of [3 x 4])
        op_mean = [np.mean(op[i], axis=2) for i in range(n_condition)]

    else:
        raise AssertionError('Please provide experiment.csv or results_condition.csv and simpos.csv files!')

    # Initialize arrays
    # cond_0, zap_1                        zap_2
    # [   |      |      |     |  ]         [   |      |      |     |  ]
    # [ ori_x  ori_y  ori_z  pos ]  . . .  [ ori_x  ori_y  ori_z  pos ]  . . .
    # [   |      |      |     |  ]         [   |      |      |     |  ]

    # Orientation and position tensors with zeromean (list over conditions containing arrays of [3 x 4 x n_con_each])
    op_zeromean = [np.zeros((3, 4, n_con_each[i])) for i in range(n_condition)]

    # Transformed coil position in (x', y', z') space (list over conditions containing arrays of [n_con_each x 3])
    p_transform = [0] * n_condition

    # SVD matrices
    u = [0] * n_condition
    s = [0] * n_condition
    v = [0] * n_condition

    # pdf parameters
    pdf_paras_location = [[[] for _ in range(3)] for _ in range(n_condition)]
    pdf_paras_orientation_euler = [[[] for _ in range(3)] for _ in range(n_condition)]
    pos_names = ['xp', 'yp', 'zp']
    orientations = [0] * n_condition

    for i in range(n_condition):

        # shift data by mean
        op_zeromean[i][:, 3, :] = op[i][:, 3, :] - np.tile(op_mean[i][:, 3][:, np.newaxis], (1, n_con_each[i]))
        # OP[i][0:3,3,:] - np.tile(OP_mean[i][:, 3][:, np.newaxis], (1, n_con_each[i]))

        # perform SVD to determine principal axis
        u[i], s[i], v[i] = np.linalg.svd(op_zeromean[i][:, 3, :].T, full_matrices=True)

        # project locations on principal axis
        p_transform[i] = np.dot(op_zeromean[i][:, 3, :].T, v[i].T)

        # rotate coil orientations to mean orientation (zero-mean)
        op_zeromean[i][:, 0:3, :] = np.dot(op[i][:, 0:3, :].transpose(2, 1, 0), op_mean[i][:, 0:3]).transpose(2, 1, 0)

        # Determine quaternions or Euler angles
        orientations[i] = np.zeros((n_con_each[i], 3))
        xlabel, ylabel = '', ''

        for j in range(n_con_each[i]):
            if orientation == 'quaternions':
                orientations[i][j, :] = pynibs.rot_to_quat(op_zeromean[i][:, 0:3, j])[1:]
                ori_names = ['xq', 'yq', 'zq']
                xlabel = ['$x_q$', '$y_q$', '$z_q$']
                ylabel = ['$p(x_q)$', '$p(y_q)$', '$p(z_q)$']

            elif orientation == 'euler':
                orientations[i][j, :] = pynibs.rotation_matrix_to_euler_angles(op_zeromean[i][:, :, j]) * 180.0 / np.pi
                ori_names = ['Psi', 'Theta', 'Phi']
                xlabel = [r'$\Psi$', r'$\Theta$', r'$\Phi$']
                ylabel = [r'$p(\Psi)$', r'$p(\Theta)$', r'$p(\Phi)$']

            else:
                raise ValueError('Please set orientation parameter to either "quaternions" or "euler"')

        # fit beta pdfs to x', y' and z' locations and orientations
        for j, name in enumerate(pos_names):
            pdf_paras_location[i][j] = pygpc.fit_betapdf(p_transform[i][:, j],
                                                         BETATOL=0.05,
                                                         PUNI=0.9,
                                                         PLOT=make_pdf_plot,
                                                         VISI=0,
                                                         xlabel=xlabel[j],
                                                         ylabel=ylabel[j],
                                                         filename=os.path.join(folder_pdfplots,
                                                                               'POS_' +
                                                                               conditions_unique[i] + '_' +
                                                                               name +
                                                                               '_betafit'))

        for j, name in enumerate(ori_names):
            pdf_paras_orientation_euler[i][j] = pygpc.fit_betapdf(orientations[i][:, j],
                                                                  BETATOL=0.05,
                                                                  PUNI=0.9,
                                                                  PLOT=make_pdf_plot,
                                                                  VISI=0,
                                                                  xlabel=xlabel[j],
                                                                  ylabel=ylabel[j],
                                                                  filename=os.path.join(folder_pdfplots,
                                                                                        'ORI_' +
                                                                                        conditions_unique[i] + '_' +
                                                                                        name +
                                                                                        '_betafit'))

    return pdf_paras_location, pdf_paras_orientation_euler, op_mean, op_zeromean, v, p_transform, orientations


def test_coil_position_gpc(parameters):
    """
    Testing valid coil positions for gPC analysis

    Parameters
    ----------

    Returns
    -------

    """
    from pygpc.RandomParameter import RandomParameter

    # load subject
    subject = pynibs.load_subject(parameters["fn_subject"])

    coil_uncertainty = False
    parameters_random = OrderedDict()

    # extract random parameters
    for p in parameters:
        if isinstance(parameters[p], RandomParameter):
            parameters_random[p] = parameters[p]
            if p in ["x", "y", "z", "psi", "theta", "phi"]:
                coil_uncertainty = True

    if coil_uncertainty:

        svd_v = parameters["coil_position_mean"][0:3, 0:3]

        # Loading geometry to create Delaunay object of outer skin surface
        with h5py.File(subject.mesh[parameters["mesh_idx"]]['fn_mesh_hdf5'], 'r') as f:
            points = np.array(f['mesh/nodes/node_coord'])
            node_number_list = np.array(f['mesh/elm/node_number_list'])
            elm_type = np.array(f['mesh/elm/elm_type'])
            regions = np.array(f['mesh/elm/tag1'])
            triangles_regions = regions[elm_type == 2,] - 1000
            triangles = node_number_list[elm_type == 2, 0:3]

        triangles = triangles[triangles_regions == 5]
        surface_points = pynibs.unique_rows(np.reshape(points[triangles], (3 * triangles.shape[0], 3)))
        limits_scaling_factor = .1

        # Generate Delaunay triangulation object
        dobj = Delaunay(surface_points)

        del points, node_number_list, elm_type, regions, triangles_regions, triangles, surface_points

        # initialize parameter dictionary
        parameters_corr = dict()
        parameters_corr_idx = dict()
        coil_param = ['x', 'y', 'z', 'psi', 'theta', 'phi']

        for i, c in enumerate(coil_param):
            parameters_corr[c] = {'pdf_limits': [0, 0], 'pdf_shape': [0, 0]}
            parameters_corr_idx[i] = c

        # we need to map some indices
        dic_par_name = dict.fromkeys([0, 1, 2], "pdf_paras_location")
        dic_par_name.update(dict.fromkeys([3, 4, 5], "pdf_paras_orientation_euler"))
        dic_limit_id = {0: 0, 2: 1}  # translates from get_invalid_coil_parameters limits id to pdf_paras_* id
        dic_param_id = dict.fromkeys([0, 3], 0)
        dic_param_id.update(dict.fromkeys([1, 4], 1))
        dic_param_id.update(dict.fromkeys([2, 5], 2))

        # Determine bad parameters and correct them
        print(("Checking validity of coil position uncertainty for condition " + parameters["cond"]))
        random.seed(1)

        # gather random variables in parameter dict (constants = 0 = no variation)
        for i_rv, rv in enumerate(parameters_random):
            if rv in coil_param:
                parameters_corr[rv] = {'pdf_limits': parameters_random[rv].pdf_limits,
                                       'pdf_shape': parameters_random[rv].pdf_shape}

        bad_params = get_invalid_coil_parameters(parameters_corr,
                                                 coil_position_mean=parameters["coil_position_mean"],
                                                 svd_v=svd_v,
                                                 del_obj=dobj,
                                                 fn_coil=parameters["fn_coil"])

        # change param limits until no dipoles left inside head
        while bad_params:
            print(" > Invalid parameter found!")
            # several parameters may lead to a bad position
            # take a random parameter from these
            param_id = random.sample(list(range(len(bad_params))), 1)[0]

            # change last bad_params[-1]' limit
            param_to_change_idx = bad_params[param_id][0]
            limit_to_change_idx = dic_limit_id[bad_params[param_id][1]]  # change min or max limit from 3 to 2 limit

            # grab old limits of parameter to change
            limits_old = parameters_corr[parameters_corr_idx[param_to_change_idx]]['pdf_limits']

            # rescale limits (but change only important boundary, min or max)
            factor = (limits_old[1] - limits_old[0]) * limits_scaling_factor
            limits_temp = [limits_old[0] + factor, limits_old[1] - factor]
            limits_new = copy.deepcopy(limits_old)
            limits_new[limit_to_change_idx] = limits_temp[limit_to_change_idx]
            parameters_corr[parameters_corr_idx[param_to_change_idx]]['pdf_limits'] = limits_new

            print((' > Changing pdf_limits of {} from {} -> {}'.format(parameters_corr_idx[param_to_change_idx],
                                                                       limits_old,
                                                                       limits_new)))

            # overwrite old values
            parameters[parameters_corr_idx[param_to_change_idx]].pdf_limits = limits_new

            # repeat dipole check
            bad_params = get_invalid_coil_parameters(parameters_corr,
                                                     coil_position_mean=parameters["coil_position_mean"],
                                                     svd_v=svd_v,
                                                     del_obj=dobj,
                                                     fn_coil=parameters["fn_coil"])

    return parameters


def get_invalid_coil_parameters(param_dict, coil_position_mean, svd_v, del_obj, fn_coil,
                                fn_hdf5_coilpos=None):
    """
    Finds gpc parameter combinations, which place coil dipoles inside subjects head.
    Only endpoints (and midpoints) of the parameter ranges are examined.

    get_invalid_coil_parameters(param_dict, pos_mean, v, del_obj, fn_coil, fn_hdf5_coilpos=None)

    Parameters
    ----------
    param_dict : dict
        Dictionary containing dictionary with ``'limits'`` and ``'pdfshape'``.
        keys: ``'x'``, ``'y'``, ``'z'``, ``'psi'``, ``'theta'``, ``'phi'``.
    coil_position_mean: np.ndarray
        (3, 4) Mean coil positions and orientations.
    svd_v : np.ndarray
        (3, 3) SVD matrix V.
    del_obj : :class:`scipy.spatial.Delaunay`
        Skin surface.
    fn_coil : str
        Filename of coil .ccd file.
    fn_hdf5_coilpos : str
        Filename of .hdf5 file to save coil_pos in (incl. path and .hdf5 extension).

    Returns
    -------
    fail_params: list of int
        Index and combination of failed parameter.
    """
    # for every condition:
    results = []

    # for i in range(6):
    limits_pos_x = param_dict['x']['limits']
    limits_pos_y = param_dict['y']['limits']
    limits_pos_z = param_dict['z']['limits']
    limits_psi = param_dict['psi']['limits']
    limits_theta = param_dict['theta']['limits']
    limits_phi = param_dict['phi']['limits']

    limits_pos_x = pynibs.add_center(limits_pos_x)
    limits_pos_y = pynibs.add_center(limits_pos_y)
    limits_pos_z = pynibs.add_center(limits_pos_z)
    limits_psi = pynibs.add_center(limits_psi)
    limits_theta = pynibs.add_center(limits_theta)
    limits_phi = pynibs.add_center(limits_phi)

    temp_list = [[0, 1, 2], [0, 1, 2], [0, 1, 2], [0, 1, 2], [0, 1, 2], [0, 1, 2]]
    combinations = list(itertools.product(*temp_list))
    del temp_list

    for combination in combinations:

        # create matsimnibs
        loc_var = np.array([limits_pos_x[combination[0]],
                            limits_pos_y[combination[1]],
                            limits_pos_z[combination[2]]])
        ori_var = np.array([limits_psi[combination[3]],
                            limits_theta[combination[4]],
                            limits_phi[combination[5]]])

        mat = calc_coil_transformation_matrix(LOC_mean=coil_position_mean[0:3, 3],
                                              ORI_mean=coil_position_mean[0:3, 0:3],
                                              LOC_var=loc_var,
                                              ORI_var=ori_var,
                                              V=svd_v)

        # get dipole points for coil for actual matsimnibs
        coil_dipoles = get_coil_dipole_pos(fn_coil, mat)

        if fn_hdf5_coilpos:
            with h5py.File(fn_hdf5_coilpos, 'w') as f:
                f.create_dataset("/dipoles/", data=coil_dipoles)

            with open(os.path.splitext(fn_hdf5_coilpos)[0] + ".xdmf", 'w') as f:
                f.write('<?xml version="1.0"?>\n')
                f.write('<!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>\n')
                f.write('<Xdmf Version="2.0" xmlns:xi="http://www.w3.org/2001/XInclude">\n')
                f.write('<Domain>\n')

                # one collection grid
                f.write('<Grid\nCollectionType="Spatial"\nGridType="Collection"\nName="Collection">\n')

                f.write('<Grid Name="coil" GridType="Uniform">\n')
                f.write('<Topology NumberOfElements="' + str(len(coil_dipoles)) +
                        '" TopologyType="Polyvertex" Name="Tri">\n')
                f.write('<DataItem Format="XML" Dimensions="' + str(len(coil_dipoles)) + ' 1">\n')
                np.savetxt(f, list(range(len(coil_dipoles))), fmt='%d',
                           delimiter=' ')  # 1 2 3 4 ... N_Points
                f.write('</DataItem>\n')
                f.write('</Topology>\n')

                # nodes
                f.write('<Geometry GeometryType="XYZ">\n')
                f.write('<DataItem Format="HDF" Dimensions="' + str(len(coil_dipoles)) + ' 3">\n')
                f.write(fn_hdf5_coilpos + ':' + '/dipoles\n')
                f.write('</DataItem>\n')
                f.write('</Geometry>\n')

                f.write('</Grid>\n')
                f.write('</Grid>\n')
                f.write('</Domain>\n')
                f.write('</Xdmf>\n')

        # check hull and add to results
        results.append(pynibs.in_hull(coil_dipoles, del_obj))

    # find parameter which drives dipoles into head
    combinations = np.array(combinations)
    fail_params = []
    if np.sum(results) > 0:
        comb_idx = np.where(np.sum(results, axis=1) > 0)[0]
        params_idx = np.where(np.var(combinations[comb_idx], axis=0) ==
                              min(np.var(combinations[comb_idx], axis=0)))[0]
        for param_idx in params_idx:
            comb_idx = comb_idx[combinations[comb_idx, params_idx[0]] != 1]
            fail_params.append((param_idx, combinations[comb_idx, params_idx[0]][0]))

    return fail_params


def create_stimsite_hdf5(fn_exp, fn_hdf, conditions_selected=None, sep="_", merge_sites=False, fix_angles=False,
                         data_dict=None, conditions_ignored=None):
    """
    Reads results_conditions and creates an hdf5/xdmf pair with condition-wise centers of stimulation sites and
    coil directions as data.

    Parameters
    ----------
    fn_exp : str
        Path to results.csv.
    fn_hdf : str
        Path where to write file. Gets overridden if already existing.
    conditions_selected : str or list of str, optional
        List of conditions returned by the function, the others are omitted.
        If None, all conditions are returned.
    sep: str, default: "_"
        Separator between condition label and angle (e.g. M1_0, or M1-0).
    merge_sites : bool
        If true, only one coil center per site is generated.
    fix_angles : bool
        rename 22.5 -> 0, 0 -> -45, 67.5 -> 90, 90 -> 135.
    data_dict : dict ofnp.ndarray of float [n_stimsites] (optional), default: None
        Dictionary containing data corresponding to the stimulation sites (keys).
    conditions_ignored : str or list of str, optional
        Conditions, which are not going to be included in the plot.

    Returns
    -------
    <Files> : hdf5/xdmf file pair
        Contains information about condition-wise stimulation sites and coil directions (fn_hdf)

    Example
    -------
    .. code-block:: python

        pynibs.create_stimsite_hdf5('/exp/1/experiment_corrected.csv',
                                    '/stimsite', True, True)
    """
    assert not fn_hdf.endswith('/')

    exp = pynibs.read_csv(fn_exp)

    exp_cond = pynibs.sort_by_condition(exp, conditions_selected=conditions_selected)  # []

    # get the unique conditions in the correct order
    conds = [c['condition'][0] for c in exp_cond]

    # remove conds
    conds_temp = []
    exp_cond_temp = []

    if type(conditions_ignored) is not list:
        conditions_ignored = [conditions_ignored]

    for i_c, c in enumerate(conds):
        ignore = False
        for ci in list(conditions_ignored):
            if c == ci:
                ignore = True

        if not ignore:
            conds_temp.append(conds[i_c])
            exp_cond_temp.append(exp_cond[i_c])

    exp_cond = exp_cond_temp
    conds = conds_temp

    # hardcoded row #3 is condition
    cond_idx = np.linspace(0, len(exp_cond), 1)[:, np.newaxis]

    centers = []
    m0 = []
    m1 = []
    m2 = []

    for i_cond in range(len(exp_cond)):
        centers.append(exp_cond[i_cond]['coil_mean_matrix'][0][0:3, 3])
        m0.append(exp_cond[i_cond]['coil_mean_matrix'][0][0:3, 0])
        m1.append(exp_cond[i_cond]['coil_mean_matrix'][0][0:3, 1])
        m2.append(exp_cond[i_cond]['coil_mean_matrix'][0][0:3, 2])

    # split conds to angles and sites: M1_90 -> M1, 90
    angles = np.array([sp.split(sep)[-1] for sp in conds]).astype(np.float64)
    sites = np.array([sp.split(sep)[0] for sp in conds])
    sites_unique = np.unique(sites)

    # average the center positions of a stimulation site over all orientations
    if merge_sites:

        # generate sites dict
        centers_sites = dict()

        for site in sites_unique:
            centers_sites[site] = []

        # gather all orientations and put them to the corresponding sites
        for i_cond, site in enumerate(sites):
            centers_sites[site].append(exp_cond[i_cond]['coil_mean_matrix'][0][0:3, 3])

        # determine average position over all orientations for each site
        for site in sites_unique:
            centers_sites[site] = np.mean(np.vstack(centers_sites[site]), axis=0)

        # write it back to centers
        for i_cond, site in enumerate(sites):
            centers[i_cond] = centers_sites[site]

    centers = np.vstack(centers)
    m0 = np.vstack(m0)
    m1 = np.vstack(m1)
    m2 = np.vstack(m2)

    # enumerate sites, as paraview does not plot string array data
    sites_idx = np.array(list(range(len(sites))))[:, np.newaxis]

    angles[angles[:] == 675.] = 67.5
    angles[angles[:] == 225.] = 22.5

    if fix_angles:
        # rename wrong angle names
        angles_cor = np.copy(angles)
        angles_cor[angles == 0] = -45.
        angles_cor[angles == 22.5] = 0.
        angles_cor[angles == 67.5] = 90.
        angles_cor[angles == 90] = 135.
        angles = angles_cor

    # write hdf5 file
    if not fn_hdf.endswith('.hdf5'):
        fn_hdf += '.hdf5'
    f = h5py.File(fn_hdf, 'w')
    f.create_dataset('centers', data=centers.astype(np.float64))
    f.create_dataset('m0', data=m0.astype(np.float64))
    f.create_dataset('m1', data=m1.astype(np.float64))
    f.create_dataset('m2', data=m2.astype(np.float64))
    f.create_dataset('cond', data=np.string_(conds))  # this is a string array, not xdmf compatible
    f.create_dataset('cond_idx', data=cond_idx)
    f.create_dataset('angles', data=angles)
    f.create_dataset('sites', data=np.string_(sites))  # this is a string array, not xdmf compatible
    f.create_dataset('sites_idx', data=sites_idx)

    data = None
    if data_dict is not None:
        data = np.zeros((len(list(data_dict.keys())), 1))
        for i_data, cond in enumerate(conds):
            data[i_data, 0] = data_dict[cond]
        f.create_dataset('data', data=data)
    f.close()

    # write .xdmf file
    f = open(fn_hdf[:-4] + 'xdmf', 'w')
    fn_hdf = os.path.basename(fn_hdf)  # relative links

    # header
    f.write('<?xml version="1.0"?>\n')
    f.write('<!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>\n')
    f.write('<Xdmf Version="2.0" xmlns:xi="http://www.w3.org/2001/XInclude">\n')
    f.write('<Domain>\n')
    f.write('<Grid\nCollectionType="Spatial"\nGridType="Collection"\nName="Collection">\n')

    # one grid for coil dipole nodes...store data hdf5.
    #######################################################
    f.write('<Grid Name="stimsites" GridType="Uniform">\n')
    f.write('<Topology NumberOfElements="' + str(centers.shape[0]) +
            '" TopologyType="Polyvertex" Name="Tri">\n')
    f.write('<DataItem Format="XML" Dimensions="' + str(centers.shape[0]) + ' 1">\n')
    # f.write(hdf5_fn + ':' + path + '/triangle_number_list\n')
    np.savetxt(f, list(range(centers.shape[0])), fmt='%d', delimiter=' ')  # 1 2 3 4 ... N_Points
    f.write('</DataItem>\n')
    f.write('</Topology>\n')

    # nodes
    f.write('<Geometry GeometryType="XYZ">\n')
    f.write('<DataItem Format="HDF" Dimensions="' + str(centers.shape[0]) + ' 3">\n')
    f.write(fn_hdf + ':' + '/centers\n')
    f.write('</DataItem>\n')
    f.write('</Geometry>\n')

    # data
    # dipole magnitude
    # the 4 vectors
    for i in range(3):
        f.write('<Attribute Name="dir_' + str(i) + '" AttributeType="Vector" Center="Cell">\n')
        f.write('<DataItem Format="HDF" Dimensions="' + str(centers.shape[0]) + ' 3">\n')
        f.write(fn_hdf + ':' + '/m' + str(i) + '\n')
        f.write('</DataItem>\n')
        f.write('</Attribute>\n\n')

    #  angles
    f.write('<Attribute Name="angles" AttributeType="Scalar" Center="Cell">\n')
    f.write('<DataItem Format="HDF" Dimensions="' + str(centers.shape[0]) + ' 1">\n')
    f.write(fn_hdf + ':' + '/angles\n')
    f.write('</DataItem>\n')
    f.write('</Attribute>\n\n')

    #  data
    if data_dict is not None:
        f.write('<Attribute Name="data" AttributeType="Scalar" Center="Cell">\n')
        f.write('<DataItem Format="HDF" Dimensions="' + str(data.shape[0]) + ' 1">\n')
        f.write(fn_hdf + ':' + '/data\n')
        f.write('</DataItem>\n')
        f.write('</Attribute>\n\n')

    # site idx
    f.write('<Attribute Name="sites_idx" AttributeType="Scalar" Center="Cell">\n')
    f.write('<DataItem Format="HDF" Dimensions="' + str(centers.shape[0]) + ' 1">\n')
    f.write(fn_hdf + ':' + '/sites_idx\n')
    f.write('</DataItem>\n')
    f.write('</Attribute>\n\n')

    f.write('</Grid>\n')
    # end coil dipole data

    # footer
    f.write('</Grid>\n')
    f.write('</Domain>\n')
    f.write('</Xdmf>\n')
    f.close()


def create_stimsite_from_list(fn_hdf, poslist, datanames=None, data=None, overwrite=False):
    """
    This takes a list of matsimnibs-style coil position and orientations and creates an .hdf5 + .xdmf tuple
    for all positions.

    Centers and coil orientations are written to disk, with optional data for each coil configuration.

    Parameters
    ----------
    fn_hdf: str
        Filename for the .hdf5 file. The .xdmf is saved with the same basename.
        Folder should already exist.
    poslist: list of np.ndarray
        (4,4) Positions.
    datanames: str or list of str, optional
        Dataset names for ``data``.
    data: np.ndarray, optional
        Dataset array with shape = ``(len(poslist.pos), len(datanames())``.
    overwrite : bool, defaul: False
        Overwrite existing files.
    """
    centers = []
    m0 = []
    m1 = []
    m2 = []
    if data is not None:
        assert isinstance(data, np.ndarray)

    for lst in poslist:
        centers.append(lst[0:3, 3])
        m0.append(lst[0:3, 0])
        m1.append(lst[0:3, 1])
        m2.append(lst[0:3, 2])

    centers = np.vstack(centers)
    m0 = np.vstack(m0)
    m1 = np.vstack(m1)
    m2 = np.vstack(m2)

    write_coil_pos_hdf5(fn_hdf, centers, m0, m1, m2, datanames=datanames, data=data, overwrite=overwrite)


def create_stimsite_from_tmslist(fn_hdf, poslist, datanames=None, data=None, overwrite=False):
    """
    This takes a :py:class:simnibs.sim_struct.TMSLIST from simnibs and creates an .hdf5 + .xdmf tuple for all positions.

    Centers and coil orientations are written to disk, with optional data for each coil configuration.

    Parameters
    ----------
    fn_hdf: str
        Filename for the .hdf5 file. The .xdmf is saved with the same basename.
        Folder should already exist.
    poslist: simnibs.sim_struct.TMSLIST
        poslist.pos[*].matsimnibs have to be set.
    datanames: str or list of str, optional
        Dataset names for ``data``.
    data: np.ndarray, optional
        Dataset array with shape = ``(len(poslist.pos), len(datanames())``.
    overwrite : bool, default: False
        Overwrite existing files
    """
    centers = []
    m0 = []
    m1 = []
    m2 = []
    assert poslist.pos
    if data is not None:
        assert isinstance(data, np.ndarray)
    for pos in poslist.pos:
        assert pos.matsimnibs is not None
        pos.matsimnibs = np.array(pos.matsimnibs)
        centers.append(pos.matsimnibs[0:3, 3])
        m0.append(pos.matsimnibs[0:3, 0])
        m1.append(pos.matsimnibs[0:3, 1])
        m2.append(pos.matsimnibs[0:3, 2])

    centers = np.vstack(centers)
    m0 = np.vstack(m0)
    m1 = np.vstack(m1)
    m2 = np.vstack(m2)

    write_coil_pos_hdf5(fn_hdf, centers, m0, m1, m2, datanames=datanames, data=data, overwrite=overwrite)


def create_stimsite_from_exp_hdf5(fn_exp, fn_hdf, datanames=None, data=None, overwrite=False):
    """
    This takes an experiment.hdf5 file and creates an .hdf5 + .xdmf tuple for all coil positions for visualization.

    Parameters
    ----------
    fn_exp : str
        Path to experiment.hdf5
    fn_hdf : str
        Filename for the resulting .hdf5 file. The .xdmf is saved with the same basename.
        Folder should already exist.
    datanames : str or list of str, optional
        Dataset names for ``data``
    data : np.ndarray, optional
        Dataset array with shape = ``(len(poslist.pos), len(datanames())``.
    overwrite : bool, default: False
        Overwrite existing files.
    """
    df_stim = pd.read_hdf(fn_exp, "stim_data")

    matsimnibs = np.zeros((4, 4, df_stim.shape[0]))

    for i in range(df_stim.shape[0]):
        matsimnibs[:, :, i] = df_stim["coil_mean"].iloc[i]

    create_stimsite_from_matsimnibs(fn_hdf=fn_hdf,
                                    matsimnibs=matsimnibs,
                                    datanames=datanames,
                                    data=data,
                                    overwrite=overwrite)


def create_stimsite_from_matsimnibs(fn_hdf, matsimnibs, datanames=None, data=None, overwrite=False):
    """
    This takes a matsimnibs array and creates an .hdf5 + .xdmf tuple for all coil positions for visualization.

    Centers and coil orientations are written disk.

    Parameters
    ----------
    fn_hdf: str
        Filename for the .hdf5 file. The .xdmf is saved with the same basename.
        Folder should already exist.
    matsimnibs: np.ndarray
        (4, 4, n_pos)
        Matsimnibs matrices containing the coil orientation (x,y,z) and position (p)

        .. math::
            \\begin{bmatrix}
            | & | & | & | \\\\
            x & y & z & p \\\\
            | & | & | & | \\\\
            0 & 0 & 0 & 1 \\\\
            \\end{bmatrix}
    datanames: str or list of str, optional
        Dataset names for ``data``.
    data: np.ndarray, optional
        (len(poslist.pos), len(datanames).
    overwrite : bool, default: False
        Overwrite existing files.
    """
    matsimnibs = np.atleast_3d(matsimnibs)
    n_pos = matsimnibs.shape[2]
    centers = np.zeros((n_pos, 3))
    m0 = np.zeros((n_pos, 3))
    m1 = np.zeros((n_pos, 3))
    m2 = np.zeros((n_pos, 3))
    if data is not None:
        assert isinstance(data, np.ndarray)

    for i in range(matsimnibs.shape[2]):
        centers[i, :] = matsimnibs[0:3, 3, i]
        m0[i, :] = matsimnibs[0:3, 0, i]
        m1[i, :] = matsimnibs[0:3, 1, i]
        m2[i, :] = matsimnibs[0:3, 2, i]

    write_coil_pos_hdf5(fn_hdf, centers, m0, m1, m2, datanames=datanames, data=data, overwrite=overwrite)


def write_coil_pos_hdf5(fn_hdf, centers, m0, m1, m2, datanames=None, data=None, overwrite=False):
    """
    Creates a ``.hdf5`` + ``.xdmf`` file tuple for all coil positions.
    Coil centers and coil orientations are saved, and - optionally - data for each position if ``data`` and
    ``datanames`` are provided.

    Parameters
    ----------
    fn_hdf : str
        Filename for the .hdf5 file. The .xdmf is saved with the same basename.
        Folder should already exist.
    centers : np.ndarray of float
        (n_pos, 3) Coil positions.
    m0 : np.ndarray of float
        (n_pos, 3) Coil orientation x-axis (looking at the active (patient) side of the coil pointing to the right).
    m1 : np.ndarray of float
        (n_pos, 3) Coil orientation y-axis (looking at the active side of the coil pointing up away from the handle).
    m2 : np.ndarray of float
        (n_pos, 3) Coil orientation z-axis (looking at the active (patient) side of the coil pointing to the patient).
    datanames : str or list of str, optional
        (n_data) Dataset names for ``data``
    data : np.ndarray, optional
        (n_pos, n_data) Dataset array with (len(poslist.pos), len(datanames()).
    overwrite : bool, default: False
        Overwrite existing files.
    """
    n_pos = centers.shape[0]
    if isinstance(datanames, str):
        datanames = [datanames]

    if data is not None:
        if datanames is None:
            raise ValueError("Provide datanames= with data= argument.")
        if isinstance(datanames, str):
            datanames = [datanames]
        if len(data.shape) <= 1:
            data = np.atleast_1d(data)[:, np.newaxis]
        assert data.shape == (n_pos, len(datanames))
    if datanames is not None and data is None:
        raise ValueError("Provide data= with datanames= argument.")

    m0_reshaped = np.hstack((m0, np.zeros((n_pos, 1)))).T[:, np.newaxis, :]
    m1_reshaped = np.hstack((m1, np.zeros((n_pos, 1)))).T[:, np.newaxis, :]
    m2_reshaped = np.hstack((m2, np.zeros((n_pos, 1)))).T[:, np.newaxis, :]
    centers_reshaped = np.hstack((centers, np.ones((n_pos, 1)))).T[:, np.newaxis, :]

    matsimnibs = np.concatenate((m0_reshaped, m1_reshaped, m2_reshaped, centers_reshaped), axis=1)

    # write hdf5 file
    if not fn_hdf.endswith('.hdf5'):
        fn_hdf += '.hdf5'
    if os.path.exists(fn_hdf) and not overwrite:
        raise OSError(fn_hdf + " already exists. Set overwrite flag for create_stimsite_from_poslist.")

    with h5py.File(fn_hdf, 'w') as f:
        f.create_dataset('centers', data=centers.astype(np.float64))
        f.create_dataset('m0', data=m0.astype(np.float64))
        f.create_dataset('m1', data=m1.astype(np.float64))
        f.create_dataset('m2', data=m2.astype(np.float64))
        f.create_dataset("matsimnibs", data=matsimnibs)

        if data is not None:
            for i, col in enumerate(data.T):
                f.create_dataset('/data/' + datanames[i], data=col)

    # write .xdmf file
    with open(fn_hdf[:-4] + 'xdmf', 'w') as f:
        fn_hdf = os.path.basename(fn_hdf)  # relative links
    
        # header
        f.write('<?xml version="1.0"?>\n')
        f.write('<!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>\n')
        f.write('<Xdmf Version="2.0" xmlns:xi="http://www.w3.org/2001/XInclude">\n')
        f.write('<Domain>\n')
        f.write('<Grid CollectionType="Spatial" GridType="Collection" Name="Collection">\n')
    
        # one grid for coil dipole nodes...store data hdf5.
        #######################################################
        f.write('<Grid Name="stimsites" GridType="Uniform">\n')
        f.write(f'<Topology NumberOfElements="{centers.shape[0]}" TopologyType="Polyvertex" Name="Tri">\n')
        f.write(f'\t<DataItem Format="XML" Dimensions="{centers.shape[0]} 1">\n')
        np.savetxt(f, list(range(centers.shape[0])), fmt='\t%d', delimiter=' ')  # 1 2 3 4 ... N_Points
        f.write('\t</DataItem>\n')
        f.write('</Topology>\n\n')
    
        # nodes
        f.write('<Geometry GeometryType="XYZ">\n')
        f.write(f'\t<DataItem Format="HDF" Dimensions="{centers.shape[0]} 3">\n')
        f.write(f'\t{fn_hdf}:/centers\n')
        f.write('\t</DataItem>\n')
        f.write('</Geometry>\n\n')
    
        # data
        # dipole magnitude
        # the 4 vectors
        for i in range(3):
            f.write(f'\t\t<Attribute Name="dir_{i}" AttributeType="Vector" Center="Cell">\n')
            f.write(f'\t\t\t<DataItem Format="HDF" Dimensions="{centers.shape[0]} 3">\n')
            f.write(f'\t\t\t{fn_hdf}:/m{i}\n')
            f.write('\t\t\t</DataItem>\n')
            f.write('\t\t</Attribute>\n\n')
    
        if data is not None:
            for i, col in enumerate(data.T):
                f.write(f'\t\t<Attribute Name="{datanames[i]}" AttributeType="Scalar" Center="Cell">\n')
                f.write('\t\t\t<DataItem Format="HDF" Dimensions="' + str(centers.shape[0]) + ' 1">\n')
                f.write(f'\t\t\t{fn_hdf}:/data/{datanames[i]}\n')
                f.write('\t\t\t</DataItem>\n')
                f.write('\t\t</Attribute>\n\n')
    
        f.write('</Grid>\n')
        # end coil dipole data
    
        # footer
        f.write('</Grid>\n')
        f.write('</Domain>\n')
        f.write('</Xdmf>\n')


def sort_opt_coil_positions(fn_coil_pos_opt, fn_coil_pos, fn_out_hdf5=None, root_path="/0/0/", verbose=False,
                            print_output=False):
    """
    Sorts coil positions according to Traveling Salesman problem

    Parameters
    ----------
    fn_coil_pos_opt : str
        Name of .hdf5 file containing the optimal coil position indices
    fn_coil_pos : str
        Name of .hdf5 file containing the matsimnibs matrices of all coil positions
    fn_out_hdf5 : str
        Name of output .hdf5 file (will be saved in the same format as fn_coil_pos_opt)
    verbose : bool, default: False
        Print output messages
    print_output : bool or str, default: False
        Print output image as .png file showing optimal path

    Returns
    -------
    <file> .hdf5 file containing the sorted optimal coil position indices
    """
    from ortools.constraint_solver import routing_enums_pb2
    from ortools.constraint_solver import pywrapcp

    if verbose:
        print(f"Loading optimal coil indices from file: {fn_coil_pos_opt}")

    with_intensities = False
    with h5py.File(fn_coil_pos_opt, "r") as f:
        coil_pos_idx = f[root_path + "coil_seq"][:, 0].astype(int)
        goal_fun_val = f[root_path + "coil_seq"][:, 1]

        with_intensities = f[root_path + "coil_seq"][:].shape[1] == 3
        if with_intensities:
            intensities_opt = f[root_path + "coil_seq"][:, 2]

    if verbose:
        print(f"Loading all coil positions from file: {fn_coil_pos}")

    with h5py.File(fn_coil_pos, "r") as f:
        matsimnibs = f["matsimnibs"][:]

    matsimnibs_opt = np.zeros((4, 4, len(coil_pos_idx)))

    for i, idx in enumerate(coil_pos_idx):
        matsimnibs_opt[:, :, i] = matsimnibs[:, :, idx]

    opt_idx_sort = np.argsort(matsimnibs_opt[0, 3, :])
    matsimnibs_opt = matsimnibs_opt[:, :, opt_idx_sort]

    unique_idx = np.zeros(len(opt_idx_sort))

    pos_unique = np.unique(matsimnibs_opt[0, 3, :])

    for i, pos in enumerate(pos_unique):
        unique_idx[matsimnibs_opt[0, 3, :] == pos] = i

    if verbose:
        print(f"Sorting same positions according to angles ...")

    matsimnibs_opt_sorted_pert = np.zeros(matsimnibs_opt.shape)
    matsimnibs_opt_sorted = np.zeros(matsimnibs_opt.shape)

    for i in unique_idx:
        pos = matsimnibs_opt[:, :, unique_idx == i]
        angles = np.zeros(pos.shape[2])

        if pos.shape[2] > 1:
            for i_p in range(pos.shape[2]):
                dot_prod = np.dot(pos[:3, 0, 0], pos[:3, 0, i_p])
                if dot_prod > 1:
                    dot_prod = 1
                elif dot_prod < -1:
                    dot_prod = -1
                angles[i_p] = np.arccos(dot_prod) / np.pi * 180

            angles_sort_idx = np.argsort(angles)
            pos_sorted = pos[:, :, angles_sort_idx]
        else:
            pos_sorted = pos

        matsimnibs_opt_sorted[:, :, unique_idx == i] = pos_sorted
        matsimnibs_opt_sorted_pert[:, :, unique_idx == i] = pos_sorted
        matsimnibs_opt_sorted_pert[0, 3, unique_idx == i] += np.arange(len(angles)) * 1e-2

    coords = matsimnibs_opt_sorted_pert[:3, 3, :].transpose()
    coords_list = [tuple(c) for c in coords]

    def compute_euclidean_distance_matrix(locations):
        """Creates callback to return distance between points."""
        distances = {}
        for from_counter, from_node in enumerate(locations):
            distances[from_counter] = {}
            for to_counter, to_node in enumerate(locations):
                if from_counter == to_counter:
                    distances[from_counter][to_counter] = 0
                else:
                    # Euclidean distance
                    distances[from_counter][to_counter] = (int(
                        math.hypot((from_node[0] - to_node[0]),
                                   (from_node[1] - to_node[1]))))
        return distances

    def get_routes(s, r, m):
        """Get vehicle routes from a solution and store them in an array."""
        # Get vehicle routes and store them in a two-dimensional array whose
        # i,j entry is the jth location visited by vehicle i along its route.
        routes_l = []
        for route_nbr in range(r.vehicles()):
            index = r.Start(route_nbr)
            route = [m.IndexToNode(index)]
            while not r.IsEnd(index):
                index = s.Value(r.NextVar(index))
                route.append(m.IndexToNode(index))
            routes_l.append(route)

        return routes_l

    def distance_callback(from_index, to_index):
        """Returns the distance between the two nodes."""
        # Convert from routing variable Index to distance matrix NodeIndex.
        from_node = manager.IndexToNode(from_index)
        to_node = manager.IndexToNode(to_index)
        return distance_matrix[from_node][to_node]

    # Instantiate the data problem.
    data = {}
    # Locations in block units
    data['locations'] = coords_list
    data['num_vehicles'] = 1
    data['depot'] = 0

    # Create the routing index manager.
    manager = pywrapcp.RoutingIndexManager(len(data['locations']), data['num_vehicles'], data['depot'])

    # Create Routing Model.
    routing = pywrapcp.RoutingModel(manager)
    distance_matrix = compute_euclidean_distance_matrix(data['locations'])
    transit_callback_index = routing.RegisterTransitCallback(distance_callback)

    # Define cost of each arc.
    routing.SetArcCostEvaluatorOfAllVehicles(transit_callback_index)

    # Setting first solution heuristic.
    search_parameters = pywrapcp.DefaultRoutingSearchParameters()
    search_parameters.first_solution_strategy = routing_enums_pb2.FirstSolutionStrategy.PATH_CHEAPEST_ARC

    # Solve the problem.
    if verbose:
        print(f"Determine optimal order of sequence ...")
    solution = routing.SolveWithParameters(search_parameters)

    routes = np.array(get_routes(solution, routing, manager)[0])[:-1]

    matsimnibs_opt_sorted = matsimnibs_opt_sorted[:, :, routes]

    idx_global_sorted = np.zeros(len(goal_fun_val))
    goal_fun_val_sorted = np.zeros(len(goal_fun_val))
    if with_intensities:
        intensities_opt_sorted = np.zeros(len(intensities_opt))

    for i in range(len(idx_global_sorted)):
        for j in range(matsimnibs.shape[2]):
            if (matsimnibs[:, :, j] == matsimnibs_opt_sorted[:, :, i]).all():
                idx_global_sorted[i] = int(j)
        goal_fun_val_sorted[i] = goal_fun_val[np.where(int(idx_global_sorted[i]) == coil_pos_idx)[0][0]]
        if with_intensities:
            intensities_opt_sorted[i] = intensities_opt[np.where(int(idx_global_sorted[i]) == coil_pos_idx)[0][0]]

    # overwrite input file if no output file given
    outpath = fn_coil_pos_opt
    if fn_out_hdf5 is not None:
        outpath = fn_out_hdf5
        shutil.copy(fn_coil_pos_opt, fn_out_hdf5)

    if verbose:
        print(f"Saving output to: {outpath}")

    with h5py.File(outpath, "a") as f:
        if with_intensities:
            f[root_path + "coil_seq"][:] = np.vstack((idx_global_sorted, goal_fun_val_sorted,
                                                      intensities_opt_sorted)).transpose()
        else:
            f[root_path + "coil_seq"][:] = np.vstack((idx_global_sorted, goal_fun_val_sorted)).transpose()

    if print_output:
        fig = plt.figure()
        ax = fig.add_subplot(projection='3d')
        if with_intensities:
            ax.scatter(coords[routes, 0], coords[routes, 1], coords[routes, 2], s=intensities_opt_sorted * 11, c="k")
        else:
            ax.scatter(coords[routes, 0], coords[routes, 1], coords[routes, 2], c="k")
        ax.plot3D(coords[routes, 0], coords[routes, 1], coords[routes, 2])
        ax.set_xlabel("x in mm")
        ax.set_ylabel("y in mm")
        ax.set_zlabel("z in mm")
        ax.view_init(elev=40., azim=230)
        plt.savefig(print_output, dpi=600, transparent=True)


def random_walk_coil(start_mat, n_steps, fn_mesh_hdf5, angles_dev=3, distance_low=1, distance_high=4,
                     angles_metric='deg', coil_pos_fn=None):
    """
    Computes random walk coil positions/orientations for a SimNIBS matsimnibs coil pos/ori.

    Parameters
    ----------
    start_mat : np.ndarry
        (4, 4) SimNIBS matsimnibs.
    n_steps : int
        Number of steps to walk.
    fn_mesh_hdf5 : str
        .hdf5 mesh filename, used to compute skin-coil distances.
    angles_dev : float or list of float, default: 3
        Angles deviation,`` np.random.normal(scale=angles_dev)``. If list, angles_dev = [alpha, beta, theta].
    distance_low : float, default: 1
        Minimum skin-coil distance.
    distance_high : float, default: 4
        Maximum skin-coil distance.
    angles_metric : str, default: 'deg'
        One of ``('deg', 'rad')``.
    coil_pos_fn : str, optional
        If provided, .hdf5/.xdmf tuple is written with coil positions/orientations.

    Returns
    -------
    walked_coils : np.ndarray
        (4, 4, n_steps + 1) coil positions / orientations.
    <file> : .hdf5/.xdmf file tupel with n_steps + 1 coil positions/orientations.
    """
    angles_dev = np.atleast_1d(np.squeeze(np.array(angles_dev)))
    if angles_dev.shape[0] == 1:
        angles_dev = np.repeat(angles_dev, 3)
    assert angles_dev.shape == (3,)

    mats = [start_mat]
    for i in range(n_steps):
        # walk position
        mat = mats[-1].copy()
        mat[:3, 3] = np.random.normal(loc=mat[:3, 3], scale=.3)

        mat_rot = start_mat.copy()

        # walk angles
        for idx, axis in enumerate(['x', 'y', 'z']):
            if angles_dev[idx] != 0:
                mat_rot = pynibs.rotate_matsimnibs_euler(axis=axis,
                                                         angle=np.random.normal(scale=angles_dev[idx]),
                                                         matsimnibs=mat_rot,
                                                         metric=angles_metric)
            mat[:3, :3] = mat_rot[:3, :3]
        mats.append(mat)

    mats = np.moveaxis(np.array(mats), 0, -1)

    # walk skin-coil distance
    assert distance_low >= 0 and distance_high >= 0
    assert distance_high >= distance_low
    distances = np.random.uniform(low=distance_low, high=distance_high, size=n_steps + 1)
    mats = pynibs.coil_distance_correction_matsimnibs(matsimnibs=mats,
                                                      fn_mesh_hdf5=fn_mesh_hdf5,
                                                      distance=distances)

    if coil_pos_fn is not None:
        pynibs.write_coil_pos_hdf5(fn_hdf=os.path.splitext(coil_pos_fn)[0],
                                   centers=mats[:3, 3, :].T,
                                   m0=mats[:3, 0, :].T,
                                   m1=mats[:3, 1, :].T,
                                   m2=mats[:3, 2, :].T,
                                   overwrite=True)

        pynibs.write_coil_sequence_xdmf(coil_pos_fn,
                                        np.arange(n_steps + 1),
                                        vec1=mats[:3, 0, :].T,
                                        vec2=mats[:3, 1, :].T,
                                        vec3=mats[:3, 2, :].T,
                                        output_xdmf=f"{os.path.splitext(coil_pos_fn)[0]}.xdmf")

    return mats
